#-------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   02/29/2008
#
#-------------------------------------------------------------------------

""" Defines the adapter classes associated with the Traits UI TabularEditor.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import (
    Any,
    Bool,
    Color,
    Either,
    Enum,
    Event,
    Float,
    Font,
    HasPrivateTraits,
    HasTraits,
    Instance,
    Int,
    Interface,
    List,
    Property,
    Str,
    cached_property,
    on_trait_change,
    provides)

#-------------------------------------------------------------------------
#  'ITabularAdapter' interface:
#-------------------------------------------------------------------------


class ITabularAdapter(Interface):

    # The row index of the current item being adapted:
    row = Int

    # The current column id being adapted (if any):
    column = Any

    # Current item being adapted:
    item = Any

    # The current value (if any):
    value = Any

    # The list of columns the adapter supports. The items in the list have the
    # same format as the *columns* trait in the *TabularAdapter* class, with the
    # additional requirement that the *string* values must correspond to a
    # *string* value in the associated *TabularAdapter* class.
    columns = List(Str)

    # Does the adapter know how to handle the current *item* or not:
    accepts = Bool

    # Does the value of *accepts* depend only upon the type of *item*?
    is_cacheable = Bool

#-------------------------------------------------------------------------
#  'AnITabularAdapter' class:
#-------------------------------------------------------------------------


@provides(ITabularAdapter)
class AnITabularAdapter(HasPrivateTraits):

    #-- Implementation of the ITabularAdapter Interface ----------------------

    # The row index of the current item being adapted:
    row = Int

    # The current column id being adapted (if any):
    column = Any

    # Current item being adapted:
    item = Any

    # The current value (if any):
    value = Any

    # The list of columns the adapter supports. The items in the list have the
    # same format as the *columns* trait in the *TabularAdapter* class, with the
    # additional requirement that the *string* values must correspond to a
    # *string* value in the associated *TabularAdapter* class.
    columns = List(Str)

    # Does the adapter know how to handle the current *item* or not:
    accepts = Bool(True)

    # Does the value of *accepts* depend only upon the type of *item*?
    is_cacheable = Bool(True)

#-------------------------------------------------------------------------
#  'TabularAdapter' class:
#-------------------------------------------------------------------------


class TabularAdapter(HasPrivateTraits):
    """ The base class for adapting list items to values that can be edited
        by a TabularEditor.
    """

    #-- Public Trait Definitions ---------------------------------------------

    # A list of columns that should appear in the table. Each entry can have one
    # of two forms: string or ( string, any ), where *string* is the UI name of
    # the column, and *any* is a value that identifies that column to the
    # adapter. Normally this value is either a trait name or an index, but it
    # can be any value that the adapter wants. If only *string* is specified,
    # then *any* is the index of the *string* within *columns*.
    columns = List()

    # Maps UI name of column to value identifying column to the adapter, if
    # different.
    column_dict = Property()

    # Specifies the default value for a new row:
    default_value = Any('')

    # The default text color for table rows (even, odd, any rows):
    odd_text_color = Color(None, update=True)
    even_text_color = Color(None, update=True)
    default_text_color = Color(None, update=True)

    # The default background color for table rows (even, odd, any rows):
    odd_bg_color = Color(None, update=True)
    even_bg_color = Color(None, update=True)
    default_bg_color = Color(None, update=True)

    # Alignment to use for a specified column:
    alignment = Enum('left', 'center', 'right')

    # The Python format string to use for a specified column:
    format = Str('%s')

    # Width of a specified column:
    width = Float(-1)

    # Can the text value of each item be edited:
    can_edit = Bool(True)

    # The value to be dragged for a specified row item:
    drag = Property

    # Can any arbitrary value be dropped onto the tabular view:
    can_drop = Bool(False)

    # Specifies where a dropped item should be placed in the table relative to
    # the item it is dropped on:
    dropped = Enum('after', 'before')

    # The font for a row item:
    font = Font(None)

    # The text color for a row item:
    text_color = Property

    # The background color for a row item:
    bg_color = Property

    # The name of the default image to use for column items:
    image = Str(None, update=True)

    # The text of a row/column item:
    text = Property

    # The content of a row/column item (may be any Python value):
    content = Property

    # The tooltip information for a row/column item:
    tooltip = Str

    # The context menu for a row/column item:
    menu = Any

    # The context menu for column header:
    column_menu = Any

    # List of optional delegated adapters:
    adapters = List(ITabularAdapter, update=True)

    #-- Traits Set by the Editor ---------------------------------------------

    # The object whose trait is being edited:
    object = Instance(HasTraits)

    # The name of the trait being edited:
    name = Str

    # The row index of the current item being adapted:
    row = Int

    # The column index of the current item being adapted:
    column = Int

    # The current column id being adapted (if any):
    column_id = Any

    # Current item being adapted:
    item = Any

    # The current value (if any):
    value = Any

    #-- Private Trait Definitions --------------------------------------------

    # Cache of attribute handlers:
    cache = Any({})

    # Event fired when the cache is flushed:
    cache_flushed = Event(update=True)

    # The mapping from column indices to column identifiers (defined by the
    # *columns* trait):
    column_map = Property(depends_on='columns')

    # The mapping from column indices to column labels (defined by the *columns*
    # trait):
    label_map = Property(depends_on='columns')

    # The name of the trait on a row item containing the value to use
    # as a row label. If None, the label will be the empty string.
    row_label_name = Either(None, Str)

    # For each adapter, specifies the column indices the adapter handles:
    adapter_column_indices = Property(depends_on='adapters,columns')

    # For each adapter, specifies the mapping from column index to column id:
    adapter_column_map = Property(depends_on='adapters,columns')

    #### TabularAdapter interface ####

    def cleanup(self):
        """ Clean up the adapter to remove references to objects.
        """
        self.trait_setq(
            object=None,
            item=None,
            value=None,
        )

    #-- Adapter methods that are sensitive to item type ----------------------

    def get_alignment(self, object, trait, column):
        """ Returns the alignment style to use for a specified column.
        """
        return self._result_for('get_alignment', object, trait, 0, column)

    def get_width(self, object, trait, column):
        """ Returns the width to use for a specified column.
        """
        return self._result_for('get_width', object, trait, 0, column)

    def get_can_edit(self, object, trait, row):
        """ Returns whether the user can edit a specified
            *object.trait[row]* item. A True result indicates the value
            can be edited, while a False result indicates that it cannot be
            edited.
        """
        return self._result_for('get_can_edit', object, trait, row, 0)

    def get_drag(self, object, trait, row):
        """ Returns the 'drag' value for a specified
            *object.trait[row]* item. A result of *None* means that the
            item cannot be dragged.
        """
        return self._result_for('get_drag', object, trait, row, 0)

    def get_can_drop(self, object, trait, row, value):
        """ Returns whether the specified *value* can be dropped on the
            specified *object.trait[row]* item. A value of **True** means the
            *value* can be dropped; and a value of **False** indicates that it
            cannot be dropped.
        """
        return self._result_for('get_can_drop', object, trait, row, 0, value)

    def get_dropped(self, object, trait, row, value):
        """ Returns how to handle a specified *value* being dropped on a
            specified *object.trait[row]* item. The possible return values are:

            'before'
                Insert the specified *value* before the dropped on item.
            'after'
                Insert the specified *value* after the dropped on item.
        """
        return self._result_for('get_dropped', object, trait, row, 0, value)

    def get_font(self, object, trait, row, column=0):
        """ Returns the font for a specified *object.trait[row]* item. A result
            of None means use the default font.
        """
        return self._result_for('get_font', object, trait, row, column)

    def get_text_color(self, object, trait, row, column=0):
        """ Returns the text color for a specified *object.trait[row]*
            item. A result of None means use the default text color.
        """
        return self._result_for('get_text_color', object, trait, row, column)

    def get_bg_color(self, object, trait, row, column=0):
        """ Returns the background color for a specified *object.trait[row]*
            item. A result of None means use the default background color.
        """
        return self._result_for('get_bg_color', object, trait, row, column)

    def get_image(self, object, trait, row, column):
        """ Returns the name of the image to use for a specified
            *object.trait[row].column* item. A result of None means no image
            should be used. Otherwise, the result should either be the name of
            the image, or an ImageResource item specifying the image to use.
        """
        return self._result_for('get_image', object, trait, row, column)

    def get_format(self, object, trait, row, column):
        """ Returns the Python format string to use for a specified column.
        """
        return self._result_for('get_format', object, trait, row, column)

    def get_text(self, object, trait, row, column):
        """ Returns the text to display for a specified
            *object.trait[row].column* item.
        """
        return self._result_for('get_text', object, trait, row, column)

    def get_content(self, object, trait, row, column):
        """ Returns the content to display for a specified
            *object.trait[row].column* item.
        """
        return self._result_for('get_content', object, trait, row, column)

    def set_text(self, object, trait, row, column, text):
        """ Sets the text for a specified *object.trait[row].column* item to
            *text*.
        """
        self._result_for('set_text', object, trait, row, column, text)

    def get_tooltip(self, object, trait, row, column):
        """ Returns the tooltip for a specified row.
        """
        return self._result_for('get_tooltip', object, trait, row, column)

    def get_menu(self, object, trait, row, column):
        """ Returns the context menu for a specified cell.
        """
        return self._result_for('get_menu', object, trait, row, column)

    def get_column_menu(self, object, trait, row, column):
        """ Returns the context menu for a specified column.
        """
        return self._result_for('get_column_menu', object, trait, row, column)

    #-- Adapter methods that are not sensitive to item type ------------------

    def get_item(self, object, trait, row):
        """ Returns the value of the *object.trait[row]* item.
        """
        try:
            return getattr(object, trait)[row]
        except:
            return None

    def len(self, object, trait):
        """ Returns the number of items in the specified *object.trait* list.
        """
        # Sometimes, during shutdown, the object has been set to None.
        if object is None:
            return 0
        else:
            return len(getattr(object, trait))

    def get_default_value(self, object, trait):
        """ Returns a new default value for the specified *object.trait* list.
        """
        return self.default_value

    def delete(self, object, trait, row):
        """ Deletes the specified *object.trait[row]* item.
        """
        del getattr(object, trait)[row]

    def insert(self, object, trait, row, value):
        """ Inserts a new value at the specified *object.trait[row]* index.
        """
        getattr(object, trait)[row: row] = [value]

    def get_column(self, object, trait, index):
        """ Returns the column id corresponding to a specified column index.
        """
        self.object, self.name = object, trait
        return self.column_map[index]

    #-- Property Implementations ---------------------------------------------

    def _get_drag(self):
        return self.item

    def _get_text_color(self):
        if (self.row % 2) == 1:
            return self.even_text_color_ or self.default_text_color

        return self.odd_text_color or self.default_text_color_

    def _get_bg_color(self):
        if (self.row % 2) == 1:
            return self.even_bg_color_ or self.default_bg_color_

        return self.odd_bg_color or self.default_bg_color_

    def _get_text(self):
        return self.get_format(
            self.object, self.name, self.row, self.column) % self.get_content(
            self.object, self.name, self.row, self.column)

    def _set_text(self, value):
        if isinstance(self.column_id, int):
            self.item[self.column_id] = self.value
        else:
            # Convert value to the correct trait type.
            try:
                trait_handler = self.item.trait(self.column_id).handler
                setattr(self.item, self.column_id,
                        trait_handler.evaluate(self.value))
            except:
                setattr(self.item, self.column_id, value)

    def _get_content(self):
        if isinstance(self.column_id, int):
            return self.item[self.column_id]

        return getattr(self.item, self.column_id)

    #-- Property Implementations ---------------------------------------------

    @cached_property
    def _get_column_dict(self):
        cols = {}
        for i, value in enumerate(self.columns):
            if isinstance(value, basestring):
                cols.update({value: value})
            else:
                cols.update({value[0]: value[1]})
        return cols

    @cached_property
    def _get_column_map(self):
        map = []
        for i, value in enumerate(self.columns):
            if isinstance(value, basestring):
                map.append(i)
            else:
                map.append(value[1])

        return map

    def get_label(self, section, obj=None):
        """Override this method if labels will vary from object to object."""
        return self.label_map[section]

    def get_row_label(self, section, obj=None):
        if self.row_label_name is None:
            return None
        rows = getattr(obj, self.name, None)
        if rows is None:
            return None
        item = rows[section]
        return getattr(item, self.row_label_name, None)

    @cached_property
    def _get_label_map(self):
        map = []
        for i, value in enumerate(self.columns):
            if isinstance(value, basestring):
                map.append(value)
            else:
                map.append(value[0])

        return map

    @cached_property
    def _get_adapter_column_indices(self):
        labels = self.label_map
        map = []
        for adapter in self.adapters:
            indices = []
            for label in adapter.columns:
                if not isinstance(label, basestring):
                    label = label[0]

                indices.append(labels.index(label))
            map.append(indices)
        return map

    @cached_property
    def _get_adapter_column_map(self):
        labels = self.label_map
        map = []
        for adapter in self.adapters:
            mapping = {}
            for label in adapter.columns:
                id = None
                if not isinstance(label, basestring):
                    label, id = label

                key = labels.index(label)
                if id is None:
                    id = key

                mapping[key] = id

            map.append(mapping)

        return map

    #-- Private Methods ------------------------------------------------------

    def _result_for(self, name, object, trait, row, column, value=None):
        """ Returns/Sets the value of the specified *name* attribute for the
            specified *object.trait[row].column* item.
        """
        self.object = object
        self.name = trait
        self.row = row
        self.column = column
        self.column_id = column_id = self.column_map[column]
        self.value = value
        self.item = item = self.get_item(object, trait, row)
        item_class = item.__class__
        key = '%s:%s:%d' % (item_class.__name__, name, column)
        handler = self.cache.get(key)
        if handler is not None:
            return handler()

        prefix = name[:4]
        trait_name = name[4:]

        for i, adapter in enumerate(self.adapters):
            if column in self.adapter_column_indices[i]:
                adapter.row = row
                adapter.item = item
                adapter.value = value
                adapter.column = column_id = self.adapter_column_map[i][column]
                if adapter.accepts:
                    get_name = '%s_%s' % (column_id, trait_name)
                    if adapter.trait(get_name) is not None:
                        if prefix == 'get_':
                            handler = lambda: getattr(adapter.set(
                                row=self.row, column=column_id,
                                item=self.item), get_name)
                        else:
                            handler = lambda: setattr(adapter.set(
                                row=self.row, column=column_id,
                                item=self.item), get_name, self.value)

                        if adapter.is_cacheable:
                            break

                        return handler()
        else:
            if item is not None and hasattr(item_class, '__mro__'):
                for klass in item_class.__mro__:
                    handler = (
                        self._get_handler_for(
                            '%s_%s_%s' %
                            (klass.__name__, column_id, trait_name), prefix) or self._get_handler_for(
                            '%s_%s' %
                            (klass.__name__, trait_name), prefix))
                    if handler is not None:
                        break

            if handler is None:
                handler = (
                    self._get_handler_for(
                        '%s_%s' %
                        (column_id, trait_name), prefix) or self._get_handler_for(
                        trait_name, prefix))

        self.cache[key] = handler
        return handler()

    def _get_handler_for(self, name, prefix):
        """ Returns the handler for a specified trait name (or None if not
            found).
        """
        if self.trait(name) is not None:
            if prefix == 'get_':
                return lambda: getattr(self, name)

            return lambda: setattr(self, name, self.value)

        return None

    @on_trait_change('columns,adapters.+update')
    def _flush_cache(self):
        """ Flushes the cache when the columns or any trait on any adapter
            changes.
        """
        self.cache = {}
        self.cache_flushed = True



""" A view containing a main walter canvas. """


# Enthought library imports.
from pyface.workbench.api import View, WorkbenchWindow
from traits.api import HasTraits, Instance, Str, on_trait_change
from traitsui.api import View as TraitsView


class DebugViewModel(HasTraits):
    """ The model for the debug view! """

    #### 'Model' interface ####################################################

    active_editor = Str
    active_part   = Str
    active_view   = Str

    window = Instance(WorkbenchWindow)

    ###########################################################################
    # 'Model' interface.
    ###########################################################################

    @on_trait_change(
        'window.active_editor', 'window.active_part', 'window.active_view'
    )
    def refresh(self):
        """ Refresh the model. """

        self.active_editor = self._get_id(self.window.active_editor)
        self.active_part   = self._get_id(self.window.active_part)
        self.active_view   = self._get_id(self.window.active_view)

        return

    def _window_changed(self):
        """ Window changed! """

        self.refresh()

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_id(self, obj):
        """ Return the Id of an object. """

        if obj is None:
            id = 'None'

        else:
            id = obj.id

        return id


class DebugView(View):
    """ A view containing a main walter canvas. """

    #### 'IWorkbenchPart' interface ###########################################

    # The part's name (displayed to the user).
    name = 'Debug'

    #### 'DebugView' interface ################################################

    # The model for the debug view!
    model = Instance(DebugViewModel)

    ###########################################################################
    # 'IWorkbenchPart' interface.
    ###########################################################################

    def create_control(self, parent):
        """ Creates the toolkit-specific control that represents the view.

        'parent' is the toolkit-specific control that is the view's parent.

        """

        self.model = DebugViewModel(window=self.window)

        ui = self.model.edit_traits(
            parent = parent,
            kind   = 'subpanel',
            view   =  TraitsView('active_part', 'active_editor', 'active_view')
        )

        return ui.control

#### EOF ######################################################################




from nose.tools import assert_equals

from traitsui.api import DefaultOverride, EditorFactory
from traits.api import HasTraits, Int


class DummyEditor(EditorFactory):

    x = Int(10)
    y = Int(20)

    def simple_editor(self, ui, object, name, description, parent):
        return ('simple_editor', self, ui, object, name, description, parent)

    def custom_editor(self, ui, object, name, description, parent):
        return ('custom_editor', self, ui, object, name, description, parent)

    def text_editor(self, ui, object, name, description, parent):
        return ('text_editor', self, ui, object, name, description, parent)

    def readonly_editor(self, ui, object, name, description, parent):
        return ('readonly_editor', self, ui, object, name, description, parent)


class NewInt(Int):

    def create_editor(self):
        return DummyEditor()


class Dummy(HasTraits):
    x = NewInt()


dummy_object = Dummy()
do = DefaultOverride(x=15, y=25, format_str='%r')


def test_simple_override():
    editor_name, editor, ui, obj, name, description, parent = do.simple_editor(
        'ui', dummy_object, 'x', 'description', 'parent')
    assert_equals(editor_name, 'simple_editor')
    assert_equals(editor.x, 15)
    assert_equals(editor.y, 25)
    assert_equals(obj, dummy_object)
    assert_equals(name, 'x')
    assert_equals(description, 'description')
    assert_equals(parent, 'parent')


def test_text_override():
    editor_name, editor, ui, obj, name, description, parent = do.text_editor(
        'ui', dummy_object, 'x', 'description', 'parent')
    assert_equals(editor_name, 'text_editor')
    assert_equals(editor.x, 15)
    assert_equals(editor.y, 25)
    assert_equals(obj, dummy_object)
    assert_equals(name, 'x')
    assert_equals(description, 'description')
    assert_equals(parent, 'parent')


def test_custom_override():
    editor_name, editor, ui, obj, name, description, parent = do.custom_editor(
        'ui', dummy_object, 'x', 'description', 'parent')
    assert_equals(editor_name, 'custom_editor')
    assert_equals(editor.x, 15)
    assert_equals(editor.y, 25)
    assert_equals(obj, dummy_object)
    assert_equals(name, 'x')
    assert_equals(description, 'description')
    assert_equals(parent, 'parent')


def test_readonly_override():
    editor_name, editor, ui, obj, name, description, parent = do.readonly_editor(
        'ui', dummy_object, 'x', 'description', 'parent')
    assert_equals(editor_name, 'readonly_editor')
    assert_equals(editor.x, 15)
    assert_equals(editor.y, 25)
    assert_equals(obj, dummy_object)
    assert_equals(name, 'x')
    assert_equals(description, 'description')
    assert_equals(parent, 'parent')



#!/usr/bin/env python

"""
A simple prototype of building a kml/kmz file for Google earth from
GNOME moss files

In this version the template and pngs are embedded.

This needs additions like drawing LEs on the beach differently, not drawing
LEs that are off teh mpa, or evaporated, or ...

"""

from mako.template import Template
import zipfile, base64


def ReadLEMossFile(filename):
    """
    Reads an LE moss file,a nd returns the coordinates of the LEs
    
    NOTE: this may not be very robust!
    """
    mossfile = file(filename, 'rU')
    LEs = []
    while True:
        mossfile.readline()
        line = mossfile.readline()
        if not line: break
        LEs.append(line.split()[:2])
    return LEs

def buildkmz(mossfilename):
    BestGuess =  ReadLEMossFile(mossfilename+".ms4")
    Uncertainty = ReadLEMossFile(mossfilename+".ms6")
    Name = mossfilename

    mytemplate = Template(LE_Template)

    outfilename = mossfilename+".kmz"
    print "rendering: %s"%outfilename

    kmlstring = mytemplate.render(LEs=BestGuess, LEs_U=Uncertainty, IconName="DotIcon", name=Name)

#    kmzfile = zipfile.ZipFile(outfilename, 'w')
    kmzfile = zipfile.ZipFile(outfilename, 'w', compression=zipfile.ZIP_DEFLATED)
    kmzfile.writestr('RedDot.png', base64.b64decode(RedDotData))
    kmzfile.writestr('BlackDot.png', base64.b64decode(BlackDotData))
    kmzfile.writestr('YellowDot.png', base64.b64decode(BlackDotData))
    kmzfile.writestr(Name+".kml", kmlstring)
    kmzfile.close()       


### the template:
LE_Template = """<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
  <Document>
    <name>${name}</name>

    <Style id="RedDotIcon">
      <IconStyle>
         <scale>1</scale>
         <Icon>
            <href>RedDot.png</href>
         </Icon>
         <hotSpot x="0.5"  y="0.5" xunits="fraction" yunits="fraction"/>  
      </IconStyle>
      <LabelStyle>
         <color>00000000</color>
      </LabelStyle>
    </Style>

    <Style id="BlackDotIcon">
      <IconStyle>
         <scale>1</scale>
         <Icon>
            <href>BlackDot.png</href>
         </Icon>
         <hotSpot x="0.5"  y="0.5" xunits="fraction" yunits="fraction"/>  
      </IconStyle>
      <LabelStyle>
         <color>00000000</color>
      </LabelStyle>
    </Style>

    <Style id="YellowDotIcon">
      <IconStyle>
         <scale>1</scale>
         <Icon>
            <href>YellowDot.png</href>
         </Icon>
         <hotSpot x="0.5"  y="0.5" xunits="fraction" yunits="fraction"/>  
      </IconStyle>
      <LabelStyle>
         <color>00000000</color>
      </LabelStyle>
    </Style>

    <StyleMap id="BlackDotPair">
        <Pair>
            <key>normal</key>
                <styleUrl>#BlackDotIcon</styleUrl>
        </Pair>
        <Pair>
            <key>highlight</key>
                <styleUrl>#YellowDotIcon</styleUrl>
        </Pair>
    </StyleMap>

    <StyleMap id="RedDotPair">
        <Pair>
            <key>normal</key>
            <styleUrl>#RedDotIcon</styleUrl>
        </Pair>
        <Pair>
            <key>highlight</key>
            <styleUrl>#RedDotIcon</styleUrl>
        </Pair>
    </StyleMap>

    <Placemark>
      <name>Uncertainty</name>
      <styleUrl>#RedDotPair</styleUrl>
      <MultiGeometry>
        % for x, y in LEs_U:
                 <Point>
                   <coordinates>${x},${y}</coordinates>
                 </Point>
        % endfor
      </MultiGeometry>
    </Placemark>

    <Placemark>
      <name>Best Guess</name>
      <styleUrl>#BlackDotPair</styleUrl>
      <MultiGeometry>
        % for x, y in LEs:
                 <Point>
                   <coordinates>${x},${y}</coordinates>
                 </Point>
        % endfor
      </MultiGeometry>
    </Placemark>

 
  </Document>
</kml>
"""

# These icons (these are base64 encoded 3-pixel sized dots in a 32x32 transparent PNG)    
BlackDotData  = "iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAAABHNCSVQICAgIfAhkiAAAAAlwSFlzAAABYgAAAWIBXyfQUwAAABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAABaSURBVFiF7dWxDQAhDEPR79NNd+x765mGCSykNI6UgoLwhIvINpP1jL5eQAEFFFBAAcCbXpQk4DvH3+latR01sACfXumc8QiU/tytCGLArRqPoIACCiiggAI2js1pdbjIwXwAAAAASUVORK5CYII="
RedDotData    = "iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAAABHNCSVQICAgIfAhkiAAAAAlwSFlzAAABZAAAAWQB3ySUyAAAABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAABaSURBVFiF7daxCcAwDETRr5Dp4n2z3rnxBMKg5gtUuLD84Aq5kjBZz+jrAgQIECBAAPC2b1YV8J3TT3etJuk1rEBOr+6c8Qiq/SG5FEEfcKnGIxAgQIAAAQI2p2Rrcf2y0MYAAAAASUVORK5CYII="
YellowDotData = "iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAAABHNCSVQICAgIfAhkiAAAAAlwSFlzAAABZAAAAWQB3ySUyAAAABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAACESURBVFiF7ZUxCsMwEATnDPpO8in9KC87v8fSpkiauLKFiRHswjVXrAY0cCGJO7Pc+roBDGAAAxhgYoAotKj0SHokLSpEGaqSdH42qhr6mY060hVD57hHAo/ddmXR82zVpA6I16Hdoa4RB0T5epBq5Of/Kf9z4MJM6oABDGAAAxjgwrwBd2LRMaa/WmgAAAAASUVORK5CYII="

if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        name = sys.argv[1]
        ext = name.split('.')[-1]
        if len(ext) == 3 and ext[:2] == 'ms':
            # this looks like it has a moss extesion, so strip it.
            name = name[:-4]
    else:
        name = "tests"
        
    buildkmz(name)
    




#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" Abstract base class for all preference pages. """


# Enthought library imports.
from traits.api import HasTraits


# fixme: in JFace this extends from 'DialogPage' which we don't have yet!
class PreferencePage(HasTraits):
    """ Abstract base class for all preference pages. """

    ###########################################################################
    # 'PreferencePage' interface.
    ###########################################################################

    def create_control(self, parent):
        """ Creates the toolkit-specific control for the page. """

        raise NotImplementedError

    def restore_defaults(self):
        """ Restore the default preferences. """

        pass

    def show_help_topic(self):
        """ Show the help topic for this preference page."""

        pass

#### EOF ######################################################################



#!/usr/bin/env python

import sys
import traceback
import wx

import logging
log = logging.getLogger(__name__)


class UndoHistoryPanel(wx.ListBox):
    def __init__(self, parent, task, **kwargs):
        self.task = task
        wx.ListBox.__init__(self, parent, wx.ID_ANY, **kwargs)

        self.Bind(wx.EVT_KEY_DOWN, self.on_key_down)

    def DoGetBestSize(self):
        """ Base class virtual method for sizer use to get the best size
        """
        width = 300
        height = -1
        best = wx.Size(width, height)

        # Cache the best size so it doesn't need to be calculated again,
        # at least until some properties of the window change
        self.CacheBestSize(best)

        return best

    def set_task(self, task):
        self.task = task

    def recalc_view(self):
        e = self.task.active_editor
        self.editor = e
        if e is not None:
            self.update_history()

    def refresh_view(self):
        editor = self.task.active_editor
        if editor is not None:
            if self.editor != editor:
                self.recalc_view()
            else:
                self.Refresh()

    def update_history(self):
        project = self.task.active_editor
        summary = project.document.undo_stack.history_list()
        self.Set(summary)
        index = project.document.undo_stack.insert_index
        if index > 0:
            self.SetSelection(index - 1)

    def activate_spring_tab(self):
        self.recalc_view()

    def get_notification_count(self):
        return 0

    def on_key_down(self, evt):
        key = evt.GetKeyCode()
        log.debug("evt=%s, key=%s" % (evt, key))



import os
import glob


def get_latest_file(pathspec):
    files = glob.glob(pathspec)
    if files:
        # for f in files:
        #     print os.path.getctime(f), os.path.getmtime(f), f
        newest = max(files, key=os.path.getmtime)
        return newest
    return pathspec



""" Tests for the preferences helper. """


# Standard library imports.
import time
import unittest

# Major package imports.
from pkg_resources import resource_filename

# Enthought library imports.
from apptools.preferences.api import Preferences, PreferencesHelper
from apptools.preferences.api import ScopedPreferences
from apptools.preferences.api import set_default_preferences
from traits.api import Any, Bool, HasTraits, Int, Float, List, Str, Unicode


def width_listener(obj, trait_name, old, new):

    width_listener.obj = obj
    width_listener.trait_name = trait_name
    width_listener.old = old
    width_listener.new = new

    return


def bgcolor_listener(obj, trait_name, old, new):

    bgcolor_listener.obj = obj
    bgcolor_listener.trait_name = trait_name
    bgcolor_listener.old = old
    bgcolor_listener.new = new

    return

# This module's package.
PKG = 'apptools.preferences.tests'


class PreferencesHelperTestCase(unittest.TestCase):
    """ Tests for the preferences helper. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        self.preferences = set_default_preferences(Preferences())

        # The filename of the example preferences file.
        self.example = resource_filename(PKG, 'example.ini')

        return

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        return

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_class_scope_preferences_path(self):
        """ class scope preferences path """

        p = self.preferences
        p.load(self.example)

        class AcmeUIPreferencesHelper(PreferencesHelper):
            """ A helper! """

            # The path to the preferences node that contains our preferences.
            preferences_path = 'acme.ui'

            # The traits that we want to initialize from preferences.
            bgcolor     = Str
            width       = Int
            ratio       = Float
            visible     = Bool
            description = Unicode
            offsets     = List(Int)
            names       = List(Str)

        helper = AcmeUIPreferencesHelper()
        helper.on_trait_change(bgcolor_listener)

        # Make sure the helper was initialized properly.
        self.assertEqual('blue', helper.bgcolor)
        self.assertEqual(50, helper.width)
        self.assertEqual(1.0, helper.ratio)
        self.assertEqual(True, helper.visible)
        self.assertEqual(u'acme ui', helper.description)
        self.assertEqual([1, 2, 3, 4], helper.offsets)
        self.assertEqual(['joe', 'fred', 'jane'], helper.names)

        # Make sure we can set the preference via the helper...
        helper.bgcolor = 'yellow'
        self.assertEqual('yellow', p.get('acme.ui.bgcolor'))
        self.assertEqual('yellow', helper.bgcolor)

        # ... and that the correct trait change event was fired.
        self.assertEqual(helper, bgcolor_listener.obj)
        self.assertEqual('bgcolor', bgcolor_listener.trait_name)
        self.assertEqual('blue', bgcolor_listener.old)
        self.assertEqual('yellow', bgcolor_listener.new)

        # Make sure we can set the preference via the preferences node...
        p.set('acme.ui.bgcolor', 'red')
        self.assertEqual('red', p.get('acme.ui.bgcolor'))
        self.assertEqual('red', helper.bgcolor)

        # ... and that the correct trait change event was fired.
        self.assertEqual(helper, bgcolor_listener.obj)
        self.assertEqual('bgcolor', bgcolor_listener.trait_name)
        self.assertEqual('yellow', bgcolor_listener.old)
        self.assertEqual('red', bgcolor_listener.new)

        return

    def test_instance_scope_preferences_path(self):
        """ instance scope preferences path """

        p = self.preferences
        p.load(self.example)

        class AcmeUIPreferencesHelper(PreferencesHelper):
            """ A helper! """

            # The traits that we want to initialize from preferences.
            bgcolor     = Str
            width       = Int
            ratio       = Float
            visible     = Bool
            description = Unicode
            offsets     = List(Int)
            names       = List(Str)

        helper = AcmeUIPreferencesHelper(preferences_path='acme.ui')
        helper.on_trait_change(bgcolor_listener)

        # Make sure the helper was initialized properly.
        self.assertEqual('blue', helper.bgcolor)
        self.assertEqual(50, helper.width)
        self.assertEqual(1.0, helper.ratio)
        self.assertEqual(True, helper.visible)
        self.assertEqual(u'acme ui', helper.description)
        self.assertEqual([1, 2, 3, 4], helper.offsets)
        self.assertEqual(['joe', 'fred', 'jane'], helper.names)

        # Make sure we can set the preference via the helper...
        helper.bgcolor = 'yellow'
        self.assertEqual('yellow', p.get('acme.ui.bgcolor'))
        self.assertEqual('yellow', helper.bgcolor)

        # ... and that the correct trait change event was fired.
        self.assertEqual(helper, bgcolor_listener.obj)
        self.assertEqual('bgcolor', bgcolor_listener.trait_name)
        self.assertEqual('blue', bgcolor_listener.old)
        self.assertEqual('yellow', bgcolor_listener.new)

        # Make sure we can set the preference via the preferences node...
        p.set('acme.ui.bgcolor', 'red')
        self.assertEqual('red', p.get('acme.ui.bgcolor'))
        self.assertEqual('red', helper.bgcolor)

        # ... and that the correct trait change event was fired.
        self.assertEqual(helper, bgcolor_listener.obj)
        self.assertEqual('bgcolor', bgcolor_listener.trait_name)
        self.assertEqual('yellow', bgcolor_listener.old)
        self.assertEqual('red', bgcolor_listener.new)

        return

    def test_default_values(self):
        """ default values """

        p = self.preferences

        class AcmeUIPreferencesHelper(PreferencesHelper):
            """ A helper! """

            # The path to the preferences node that contains our preferences.
            preferences_path = 'acme.ui'

            # The traits that we want to initialize from preferences.
            bgcolor     = Str('blue')
            width       = Int(50)
            ratio       = Float(1.0)
            visible     = Bool(True)
            description = Unicode(u'description')
            offsets     = List(Int, [1, 2, 3, 4])
            names       = List(Str, ['joe', 'fred', 'jane'])

        helper = AcmeUIPreferencesHelper()

        # Make sure the helper was initialized properly.
        self.assertEqual('blue', helper.bgcolor)
        self.assertEqual(50, helper.width)
        self.assertEqual(1.0, helper.ratio)
        self.assertEqual(True, helper.visible)
        self.assertEqual(u'description', helper.description)
        self.assertEqual([1, 2, 3, 4], helper.offsets)
        self.assertEqual(['joe', 'fred', 'jane'], helper.names)

        return

    def test_real_unicode_values(self):
        """ Test with real life unicode values """

        p = self.preferences
        p.load(self.example)

        class AcmeUIPreferencesHelper(PreferencesHelper):
            """ A helper! """

            # The path to the preferences node that contains our preferences.
            preferences_path = 'acme.ui'

            # The traits that we want to initialize from preferences.
            bgcolor     = Str('blue')
            width       = Int(50)
            ratio       = Float(1.0)
            visible     = Bool(True)
            description = Unicode(u'')
            offsets     = List(Int, [1, 2, 3, 4])
            names       = List(Str, ['joe', 'fred', 'jane'])

        helper = AcmeUIPreferencesHelper()

        first_unicode_str = u'U\xdc\xf2ser'
        first_utf8_str = 'U\xc3\x9c\xc3\xb2ser'

        original_description = helper.description
        helper.description = first_unicode_str
        self.assertEqual(first_unicode_str, helper.description)


        second_unicode_str = u'caf\xe9'
        second_utf8_str = 'caf\xc3\xa9'
        helper.description = second_unicode_str
        self.assertEqual(second_unicode_str, helper.description)
        self.assertEqual(second_unicode_str, p.get('acme.ui.description'))

        p.save(self.example)

        p.load(self.example)
        self.assertEqual(second_unicode_str, p.get('acme.ui.description'))
        self.assertEqual(u'True', p.get('acme.ui.visible'))
        self.assertEqual(True, helper.visible)

        # reset the original description and save the example file
        helper.description = original_description
        p.save(self.example)

    def test_no_preferences_path(self):
        """ no preferences path """

        p = self.preferences
        p.load(self.example)

        class AcmeUIPreferencesHelper(PreferencesHelper):
            """ A helper! """

            # The traits that we want to initialize from preferences.
            bgcolor     = Str
            width       = Int
            ratio       = Float
            visible     = Bool
            description = Unicode
            offsets     = List(Int)
            names       = List(Str)

        # Cannot create a helper with a preferences path.
        self.assertRaises(SystemError, AcmeUIPreferencesHelper)

        return

    def test_sync_trait(self):
        """ sync trait """

        class Widget(HasTraits):
            """ A widget! """

            background_color = Str

        w = Widget()
        w.on_trait_change(bgcolor_listener)

        p = self.preferences
        p.load(self.example)

        class AcmeUIPreferencesHelper(PreferencesHelper):
            """ A helper! """

            # The path to the preferences node that contains our preferences.
            preferences_path = 'acme.ui'

            # The traits that we want to initialize from preferences.
            bgcolor     = Str
            width       = Int
            ratio       = Float
            visible     = Bool
            description = Unicode
            offsets     = List(Int)
            names       = List(Str)

        helper = AcmeUIPreferencesHelper()
        helper.sync_trait('bgcolor', w, 'background_color')

        # Make sure the helper was initialized properly.
        self.assertEqual('blue', helper.bgcolor)
        self.assertEqual(50, helper.width)
        self.assertEqual(1.0, helper.ratio)
        self.assertEqual(True, helper.visible)
        self.assertEqual(u'acme ui', helper.description)
        self.assertEqual([1, 2, 3, 4], helper.offsets)
        self.assertEqual(['joe', 'fred', 'jane'], helper.names)

        self.assertEqual('blue', w.background_color)

        # Make sure we can set the preference via the helper...
        helper.bgcolor = 'yellow'
        self.assertEqual('yellow', p.get('acme.ui.bgcolor'))
        self.assertEqual('yellow', helper.bgcolor)

        self.assertEqual('yellow', w.background_color)

        # ... and that the correct trait change event was fired.
        self.assertEqual(w, bgcolor_listener.obj)
        self.assertEqual('background_color', bgcolor_listener.trait_name)
        self.assertEqual('blue', bgcolor_listener.old)
        self.assertEqual('yellow', bgcolor_listener.new)

        # Make sure we can set the preference via the preferences node...
        p.set('acme.ui.bgcolor', 'red')
        self.assertEqual('red', p.get('acme.ui.bgcolor'))
        self.assertEqual('red', helper.bgcolor)

        self.assertEqual('red', w.background_color)

        # ... and that the correct trait change event was fired.
        self.assertEqual(w, bgcolor_listener.obj)
        self.assertEqual('background_color', bgcolor_listener.trait_name)
        self.assertEqual('yellow', bgcolor_listener.old)
        self.assertEqual('red', bgcolor_listener.new)

        return

    def test_scoped_preferences(self):
        """ scoped preferences """

        p = set_default_preferences(ScopedPreferences())

        # Set a preference value in the default scope.
        p.set('default/acme.ui.bgcolor', 'blue')

        class AcmeUIPreferencesHelper(PreferencesHelper):
            """ A helper! """

            # The path to the preferences node that contains our preferences.
            preferences_path = 'acme.ui'

            # The traits that we want to initialize from preferences.
            bgcolor = Str

            # A trait for a preference that does not exist yet.
            name = Str

        helper = AcmeUIPreferencesHelper()

        # Make sure the trait is set!
        self.assertEqual('blue', helper.bgcolor)

        # And that the non-existent trait gets the default value.
        self.assertEqual('', helper.name)

        return

    def test_preference_not_in_file(self):
        """ preference not in file """

        class AcmeUIPreferencesHelper(PreferencesHelper):
            """ A helper! """

            # The path to the preferences node that contains our preferences.
            preferences_path = 'acme.ui'

            # A trait that has no corresponding value in the file.
            title = Str('Acme')

        helper = AcmeUIPreferencesHelper()

        # Make sure the trait is set!
        self.assertEqual('Acme', helper.title)

        # Set a new value.
        helper.title = 'Acme Plus'

        # Make sure the trait is set!
        self.assertEqual('Acme Plus', helper.title)
        self.assertEqual('Acme Plus', self.preferences.get('acme.ui.title'))

        return

    def test_preferences_node_changed(self):
        """ preferences node changed """

        p = self.preferences
        p.load(self.example)

        class AcmeUIPreferencesHelper(PreferencesHelper):
            """ A helper! """

            # The path to the preferences node that contains our preferences.
            preferences_path = 'acme.ui'

            # The traits that we want to initialize from preferences.
            bgcolor     = Str
            width       = Int
            ratio       = Float
            visible     = Bool
            description = Unicode
            offsets     = List(Int)
            names       = List(Str)

        helper = AcmeUIPreferencesHelper()

        # We only listen to some of the traits so the testing is easier.
        helper.on_trait_change(width_listener, ['width'])
        helper.on_trait_change(bgcolor_listener, ['bgcolor'])

        # Create a new preference node.
        p1 = Preferences()
        p1.load(self.example)

        p1.set('acme.ui.bgcolor', 'red')
        p1.set('acme.ui.width', 40)

        # Set the new preferences
        helper.preferences = p1

        # Test event handling.
        self.assertEqual(helper, width_listener.obj)
        self.assertEqual('width', width_listener.trait_name)
        self.assertEqual(50, width_listener.old)
        self.assertEqual(40, width_listener.new)

        self.assertEqual(helper, bgcolor_listener.obj)
        self.assertEqual('bgcolor', bgcolor_listener.trait_name)
        self.assertEqual('blue', bgcolor_listener.old)
        self.assertEqual('red', bgcolor_listener.new)

        # Test re-initialization.
        self.assertEqual(helper.bgcolor, 'red')
        self.assertEqual(helper.width, 40)

        # Test event handling.
        p1.set('acme.ui.bgcolor', 'black')
        self.assertEqual(helper, bgcolor_listener.obj)
        self.assertEqual('bgcolor', bgcolor_listener.trait_name)
        self.assertEqual('red', bgcolor_listener.old)
        self.assertEqual('black', bgcolor_listener.new)

        # This should not trigger any new changes since we are setting values
        # on the old preferences node.
        p.set('acme.ui.bgcolor', 'white')
        self.assertEqual(helper, bgcolor_listener.obj)
        self.assertEqual('bgcolor', bgcolor_listener.trait_name)
        self.assertEqual('red', bgcolor_listener.old)
        self.assertEqual('black', bgcolor_listener.new)

        return

    def test_nested_set_in_trait_change_handler(self):
        """ nested set in trait change handler """

        p = self.preferences
        p.load(self.example)

        class AcmeUIPreferencesHelper(PreferencesHelper):
            """ A helper! """

            # The traits that we want to initialize from preferences.
            bgcolor     = Str
            width       = Int
            ratio       = Float
            visible     = Bool
            description = Unicode
            offsets     = List(Int)
            names       = List(Str)

            # When the width changes, change the ratio.
            def _width_changed(self, trait_name, old, new):
                """ Static trait change handler. """

                self.ratio = 3.0

                return

        helper = AcmeUIPreferencesHelper(preferences_path='acme.ui')

        # Make sure the helper was initialized properly.
        self.assertEqual('blue', helper.bgcolor)
        self.assertEqual(50, helper.width)
        self.assertEqual(1.0, helper.ratio)
        self.assertEqual(True, helper.visible)
        self.assertEqual(u'acme ui', helper.description)
        self.assertEqual([1, 2, 3, 4], helper.offsets)
        self.assertEqual(['joe', 'fred', 'jane'], helper.names)

        # Change the width via the preferences node. This should cause the
        # ratio to get set via the static trait change handler on the helper.
        p.set('acme.ui.width', 42)
        self.assertEqual(42, helper.width)
        self.assertEqual('42', p.get('acme.ui.width'))

        # Did the ratio get changed?
        self.assertEqual(3.0, helper.ratio)
        self.assertEqual('3.0', p.get('acme.ui.ratio'))

        return

    # fixme: No comments - nice work... I added the doc string and the 'return'
    # to be compatible with the rest of the module. Interns please note correct
    # procedure when modifying existing code. If in doubt, ask a developer.
    def test_unevaluated_strings(self):
        """ unevaluated strings """

        p = self.preferences
        p.load(self.example)

        class AcmeUIPreferencesHelper(PreferencesHelper):
            width = Any(is_str=True)

        helper = AcmeUIPreferencesHelper(preferences_path='acme.ui')

        self.assertEqual('50', helper.width)

        return


# Entry point for stand-alone testing.
if __name__ == '__main__':
    unittest.main()

#### EOF ######################################################################



#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   07/14/2008
#
#-------------------------------------------------------------------------

""" Editor factory for scrubber-based integer or float value editors.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from pyface.ui_traits import Alignment
from traits.api import Float, Color, Property

from ..basic_editor_factory import BasicEditorFactory

from ..toolkit import toolkit_object

#-------------------------------------------------------------------------
#  Create the editor factory object:
#-------------------------------------------------------------------------

# Editor factory for scrubber-based integer or float value editors.


class ScrubberEditor(BasicEditorFactory):

    # The editor class to be instantiated:
    klass = Property

    # The low end of the scrubber range:
    low = Float

    # The high end of the scrubber range:
    high = Float

    # The normal increment (default: auto-calculate):
    increment = Float

    # The alignment of the text within the scrubber:
    alignment = Alignment('center')

    # The background color for the scrubber:
    color = Color(None)

    # The hover mode background color for the scrubber:
    hover_color = Color(None)

    # The active mode background color for the scrubber:
    active_color = Color(None)

    # The scrubber border color:
    border_color = Color(None)

    # The color to use for the value text:
    text_color = Color('black')

    def _get_klass(self):
        """ Returns the toolkit-specific editor class to be instantiated.
        """
        return toolkit_object('scrubber_editor:_ScrubberEditor')

### EOF ##################################################################



""" A plugin manager composed of other plugin managers! """


# Standard library imports.
import logging

# Enthought library imports.
from traits.api import Event, HasTraits, Instance, List, provides
from traits.api import on_trait_change

# Local imports.
from .i_application import IApplication
from .i_plugin import IPlugin
from .i_plugin_manager import IPluginManager
from .plugin_event import PluginEvent
from .plugin_manager import PluginManager


# Logging.
logger = logging.getLogger(__name__)


@provides(IPluginManager)
class CompositePluginManager(HasTraits):
    """ A plugin manager composed of other plugin managers!

    e.g::

        plugin_manager = CompositePluginManager(
             plugin_mangers = [
                 EggBasketPluginManager(...),
                 PackagePluginManager(...),
             ]
        )

    """

    #### 'IPluginManager' protocol #############################################

    #### Events ####

    # Fired when a plugin has been added to the manager.
    plugin_added = Event(PluginEvent)

    # Fired when a plugin has been removed from the manager.
    plugin_removed = Event(PluginEvent)

    #### 'CompositePluginManager' protocol #####################################

    # The application that the plugin manager is part of.
    application = Instance(IApplication)
    def _application_changed(self, trait_name, old, new):
        for plugin_manager in self.plugin_managers:
            plugin_manager.application = new

        return

    # The plugin managers that make up this plugin manager!
    #
    # This is currently a list of 'PluginManager's as opposed to, the more
    # preferable 'IPluginManager' because the interface doesn't currently
    # have an 'application' trait. Should we move 'application' up to
    # 'IPluginManager'?
    plugin_managers = List(PluginManager)
    @on_trait_change('plugin_managers[]')
    def _update_application(self, obj, trait_named, removed, added):
        for plugin_manager in removed:
            plugin_manager.application = self.application

        for plugin_manager in added:
            plugin_manager.application = self.application

    @on_trait_change('plugin_managers:plugin_added')
    def _plugin_added(self, obj, trait_name, old, new):
        self.plugin_added = new

    @on_trait_change('plugin_managers:plugin_removed')
    def _plugin_removed(self, obj, trait_name, old, new):
        self.plugin_removed = new

    #### Private protocol ######################################################

    # The plugins that the manager manages!
    _plugins = List(IPlugin)
    def __plugins_default(self):
        plugins = []
        for plugin_manager in self.plugin_managers:
            for plugin in plugin_manager:
                plugins.append(plugin)

        return plugins

    #### 'object' protocol ####################################################

    def __iter__(self):
        """ Return an iterator over the manager's plugins. """

        plugins = []
        for plugin_manager in self.plugin_managers:
            for plugin in plugin_manager:
                plugins.append(plugin)

        return iter(plugins)

    #### 'IPluginManager' protocol #############################################

    def add_plugin(self, plugin):
        """ Add a plugin to the manager. """

        raise NotImplementedError

    def get_plugin(self, plugin_id):
        """ Return the plugin with the specified Id. """

        for plugin in self:
            if plugin_id == plugin.id:
                break

        else:
            plugin = None

        return plugin

    def remove_plugin(self, plugin):
        """ Remove a plugin from the manager. """

        raise NotImplementedError

    def start(self):
        """ Start the plugin manager. """

        for plugin in self:
            self.start_plugin(plugin)

        return

    def start_plugin(self, plugin=None, plugin_id=None):
        """ Start the specified plugin. """

        plugin = plugin or self.get_plugin(plugin_id)
        if plugin is not None:
            logger.debug('plugin %s starting', plugin.id)
            plugin.activator.start_plugin(plugin)
            logger.debug('plugin %s started', plugin.id)

        else:
            raise SystemError('no such plugin %s' % plugin_id)

        return

    def stop(self):
        """ Stop the plugin manager. """

        # We stop the plugins in the reverse order that they were started.
        stop_order = list(iter(self))
        stop_order.reverse()

        for plugin in stop_order:
            self.stop_plugin(plugin)

        return

    def stop_plugin(self, plugin=None, plugin_id=None):
        """ Stop the specified plugin. """

        plugin = plugin or self.get_plugin(plugin_id)
        if plugin is not None:
            logger.debug('plugin %s stopping', plugin.id)
            plugin.activator.stop_plugin(plugin)
            logger.debug('plugin %s stopped', plugin.id)

        else:
            raise SystemError('no such plugin %s' % plugin_id)

        return

#### EOF ######################################################################



import wx

from traits.api import Instance, Int, Str
from traitsui.wx.editor import Editor
from pyface.ui.wx.progress_dialog import ProgressDialog


class _ProgressDialog(ProgressDialog):

    def close(self):
        """ Overwritten to disable closing.
        """
        pass


class SimpleEditor(Editor):
    """
    Show a progress bar with all the optional goodies

    """

    progress = Instance(ProgressDialog)

    # The message to be displayed along side the progress guage
    message = Str

    # The starting value
    min = Int

    # The ending value
    max = Int

    #-- Editor interface ------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.control = self.create_control(parent)
        factory = self.factory
        self.min = factory.min
        self.max = factory.max
        self.message = factory.message
        self.sync_value(factory.min_name, 'min', 'from')
        self.sync_value(factory.max_name, 'max', 'from')
        self.sync_value(factory.message_name, 'message', 'from')
        self.set_tooltip()

    def create_control(self, parent):
        """
        Finishes initializing the editor by creating the underlying widget.
        """

        self.progress = ProgressDialog(title=self.factory.title,
                                       message=self.factory.message,
                                       min=self.factory.min,
                                       max=self.factory.max,
                                       can_cancel=self.factory.can_cancel,
                                       show_time=self.factory.show_time,
                                       show_percent=self.factory.show_percent)

        panel = wx.Panel(parent, -1)

        sizer = wx.BoxSizer(wx.VERTICAL)
        panel.SetSizer(sizer)
        panel.SetAutoLayout(True)
        panel.SetBackgroundColour(wx.NullColor)

        self.progress.dialog_size = wx.Size()

        # The 'guts' of the dialog.
        self.progress._create_message(panel, sizer)
        self.progress._create_gauge(panel, sizer)
        self.progress._create_percent(panel, sizer)
        self.progress._create_timer(panel, sizer)
        self.progress._create_buttons(panel, sizer)

        panel.SetClientSize(self.progress.dialog_size)

        panel.CentreOnParent()

        self.control = panel
        return self.control

    def update_editor(self):
        """
        Updates the editor when the object trait changes externally to the
        editor.
        """
        self.progress.min = self.min
        self.progress.max = self.max
        self.progress.change_message(self.message)
        if self.value:
            self.progress.update(self.value)
        return

    def _min_changed(self):
        self.update_editor()

    def _max_changed(self):
        self.update_editor()

    def _message_changed(self):
        self.update_editor()



# Enthought library imports.
from envisage.ui.tasks.api import PreferencesPane, TaskFactory
from apptools.preferences.api import PreferencesHelper
from traits.api import Str, Int
from traitsui.api import HGroup, VGroup, Item, Label, \
    View


class HtmlViewPreferences(PreferencesHelper):
    """ The preferences helper for the Framework application.
    """

    #### 'PreferencesHelper' interface ########################################

    # The path to the preference node that contains the preferences.
    preferences_path = 'omnivore.tasks.htmlview'

    #### Preferences ##########################################################

    # Font name for normal (proportional spacing) rendering
    normal_face = Str("")

    # Font name for monospace rendering
    fixed_face = Str("")

    # Default font size
    font_size = Int(10)


class HtmlViewPreferencesPane(PreferencesPane):
    """ The preferences pane for the Framework application.
    """

    #### 'PreferencesPane' interface ##########################################

    # The factory to use for creating the preferences model object.
    model_factory = HtmlViewPreferences

    category = Str('Editors')

    #### 'FrameworkPreferencesPane' interface ################################

    view = View(
        VGroup(HGroup(Item('normal_face'),
                      Label('Proportional Font Name'),
                      show_labels = False),
               HGroup(Item('fixed_face'),
                      Label('Fixed Width Font Name'),
                      show_labels = False),
               HGroup(Item('font_size'),
                      Label('Default Point Size'),
                      show_labels = False),
               label='HTML Viewer'),
        resizable=True)




import sys
import unittest
import marshal
import imp
import struct
from textwrap import dedent

from fs.expose.importhook import FSImportHook
from fs.tempfs import TempFS
from fs.zipfs import ZipFS

from six import b


class TestFSImportHook(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        for mph in list(sys.meta_path):
            if isinstance(mph,FSImportHook):
                sys.meta_path.remove(mph)
        for ph in list(sys.path_hooks):
            if issubclass(ph,FSImportHook):
                sys.path_hooks.remove(mph)
        for (k,v) in sys.modules.items():
            if k.startswith("fsih_"):
                del sys.modules[k]
            elif hasattr(v,"__loader__"):
                if isinstance(v.__loader__,FSImportHook):
                    del sys.modules[k]
        sys.path_importer_cache.clear()

    def _init_modules(self,fs):
        fs.setcontents("fsih_hello.py",b(dedent("""
            message = 'hello world!'
        """)))
        fs.makedir("fsih_pkg")
        fs.setcontents("fsih_pkg/__init__.py",b(dedent("""
            a = 42
        """)))
        fs.setcontents("fsih_pkg/sub1.py",b(dedent("""
            import fsih_pkg
            from fsih_hello import message
            a = fsih_pkg.a
        """)))
        fs.setcontents("fsih_pkg/sub2.pyc",self._getpyc(b(dedent("""
            import fsih_pkg
            from fsih_hello import message
            a = fsih_pkg.a * 2
        """))))

    def _getpyc(self,src):
        """Get the .pyc contents to match th given .py source code."""
        code = imp.get_magic() + struct.pack("<i",0)
        code += marshal.dumps(compile(src,__file__,"exec"))
        return code

    def test_loader_methods(self):
        t = TempFS()
        self._init_modules(t)
        ih = FSImportHook(t)
        sys.meta_path.append(ih)
        try:
            self.assertEquals(ih.find_module("fsih_hello"),ih)
            self.assertEquals(ih.find_module("fsih_helo"),None)
            self.assertEquals(ih.find_module("fsih_pkg"),ih)
            self.assertEquals(ih.find_module("fsih_pkg.sub1"),ih)
            self.assertEquals(ih.find_module("fsih_pkg.sub2"),ih)
            self.assertEquals(ih.find_module("fsih_pkg.sub3"),None)
            m = ih.load_module("fsih_hello")
            self.assertEquals(m.message,"hello world!")
            self.assertRaises(ImportError,ih.load_module,"fsih_helo")
            ih.load_module("fsih_pkg")
            m = ih.load_module("fsih_pkg.sub1")
            self.assertEquals(m.message,"hello world!")
            self.assertEquals(m.a,42)
            m = ih.load_module("fsih_pkg.sub2")
            self.assertEquals(m.message,"hello world!")
            self.assertEquals(m.a,42 * 2)
            self.assertRaises(ImportError,ih.load_module,"fsih_pkg.sub3")
        finally:
            sys.meta_path.remove(ih)
            t.close()

    def _check_imports_are_working(self):
        try:
            import fsih_hello
            self.assertEquals(fsih_hello.message,"hello world!")
            try:
                import fsih_helo
            except ImportError:
                pass
            else:
                assert False, "ImportError not raised"
            import fsih_pkg
            import fsih_pkg.sub1
            self.assertEquals(fsih_pkg.sub1.message,"hello world!")
            self.assertEquals(fsih_pkg.sub1.a,42)
            import fsih_pkg.sub2
            self.assertEquals(fsih_pkg.sub2.message,"hello world!")
            self.assertEquals(fsih_pkg.sub2.a,42 * 2)
            try:
                import fsih_pkg.sub3
            except ImportError:
                pass
            else:
                assert False, "ImportError not raised"
        finally:
            for k in sys.modules.keys():
                if k.startswith("fsih_"):
                    del sys.modules[k]

    def test_importer_on_meta_path(self):
        t = TempFS()
        self._init_modules(t)
        ih = FSImportHook(t)
        sys.meta_path.append(ih)
        try:
            self._check_imports_are_working()
        finally:
            sys.meta_path.remove(ih)
            t.close()

    def test_url_on_sys_path(self):
        t = TempFS()
        zpath = t.getsyspath("modules.zip")
        z = ZipFS(zpath,"w")
        self._init_modules(z)
        z.close()
        z = ZipFS(zpath,"r")
        assert z.isfile("fsih_hello.py")
        z.close()
        sys.path.append("zip://" + zpath)
        FSImportHook.install()
        try:
            self._check_imports_are_working()
        finally:
            sys.path_hooks.remove(FSImportHook)
            sys.path.pop()
            t.close()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/25/2004
#
#------------------------------------------------------------------------------

""" Defines various helper functions that are useful for creating Traits-based
    user interfaces.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from operator import itemgetter

from traits.api import BaseTraitHandler, CTrait, Enum, TraitError

from .ui_traits import SequenceTypes

#-------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------

# Layout orientation for a control and its associated editor
Orientation = Enum('horizontal', 'vertical')

# Docking drag bar style:
DockStyle = Enum('horizontal', 'vertical', 'tab', 'fixed')

#----------------------------------------------------------------------------
#  Return a 'user-friendly' name for a specified trait:
#----------------------------------------------------------------------------


def user_name_for(name):
    """ Returns a "user-friendly" name for a specified trait.
    """
    name = name.replace('_', ' ')
    name = name[:1].upper() + name[1:]
    result = ''
    last_lower = 0
    for c in name:
        if c.isupper() and last_lower:
            result += ' '
        last_lower = c.islower()
        result += c
    return result

#-------------------------------------------------------------------------
#  Format a number with embedded commas:
#-------------------------------------------------------------------------


def commatize(value):
    """ Formats a specified value as an integer string with embedded commas.
        For example: commatize( 12345 ) returns "12,345".
    """
    s = str(abs(value))
    s = s.rjust(((len(s) + 2) / 3) * 3)
    result = ','.join([s[i: i + 3] for i in range(0, len(s), 3)]).lstrip()
    if value >= 0:
        return result

    return '-' + result

#-------------------------------------------------------------------------
#  Recomputes the mappings for a new set of enumeration values:
#-------------------------------------------------------------------------


def enum_values_changed(values, strfunc=unicode):
    """ Recomputes the mappings for a new set of enumeration values.
    """

    if isinstance(values, dict):
        data = [(strfunc(v), n) for n, v in values.items()]
        if len(data) > 0:
            data.sort(key=itemgetter(0))
            col = data[0][0].find(':') + 1
            if col > 0:
                data = [(n[col:], v) for n, v in data]
    elif not isinstance(values, SequenceTypes):
        handler = values
        if isinstance(handler, CTrait):
            handler = handler.handler
        if not isinstance(handler, BaseTraitHandler):
            raise TraitError("Invalid value for 'values' specified")
        if handler.is_mapped:
            data = [(strfunc(n), n) for n in handler.map.keys()]
            data.sort(key=itemgetter(0))
        else:
            data = [(strfunc(v), v) for v in handler.values]
    else:
        data = [(strfunc(v), v) for v in values]

    names = [x[0] for x in data]
    mapping = {}
    inverse_mapping = {}
    for name, value in data:
        mapping[name] = value
        inverse_mapping[value] = name

    return (names, mapping, inverse_mapping)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#------------------------------------------------------------------------------

""" Defines the various button editors for the wxPython user interface toolkit.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------


import wx

from traits.api \
    import Str

# FIXME: ToolkitEditorFactory is a proxy class defined here just for backward
# compatibility. The class has been moved to the
# traitsui.editors.button_editor file.
from traitsui.editors.button_editor \
    import ToolkitEditorFactory

from editor \
    import Editor

#-------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------


class SimpleEditor(Editor):
    """ Simple style editor for a button.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # The button label
    label = Str

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        label = self.factory.label or self.item.get_label(self.ui)
        self.control = wx.Button(parent, -1, self.string_value(label))
        self.sync_value(self.factory.label_value, 'label', 'from')
        wx.EVT_BUTTON(parent, self.control.GetId(), self.update_object)
        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Handles the 'label' trait being changed:
    #-------------------------------------------------------------------------

    def _label_changed(self, label):
        self.control.SetLabel(self.string_value(label))

    #-------------------------------------------------------------------------
    #  Handles the user clicking the button by setting the value on the object:
    #-------------------------------------------------------------------------

    def update_object(self, event):
        """ Handles the user clicking the button by setting the factory value
            on the object.
        """
        factory = self.factory
        self.value = factory.value

        # If there is an associated view, then display it:
        if factory.view is not None:
            self.object.edit_traits(view=factory.view,
                                    parent=self.control)

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        pass

    #-------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #-------------------------------------------------------------------------

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        wx.EVT_BUTTON(self.control.GetParent(), self.control.GetId(), None)

        super(SimpleEditor, self).dispose()

#-------------------------------------------------------------------------
#  'CustomEditor' class:
#-------------------------------------------------------------------------


class CustomEditor(SimpleEditor):
    """ Custom style editor for a button, which can contain an image.
    """

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        from pyface.image_button import ImageButton

        factory = self.factory
        self._control = ImageButton(
            parent,
            label=self.string_value(factory.label),
            image=factory.image,
            style=factory.style,
            orientation=factory.orientation,
            width_padding=factory.width_padding,
            height_padding=factory.height_padding
        )
        self.control = self._control.control
        self._control.on_trait_change(self.update_object, 'clicked',
                                      dispatch='ui')
        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #-------------------------------------------------------------------------

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        self._control.on_trait_change(self.update_object, 'clicked',
                                      remove=True)

        super(CustomEditor, self).dispose()

### EOF #######################################################################



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005-2013, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/07/2004
#
#------------------------------------------------------------------------------

from __future__ import absolute_import

from ._version import full_version as __version__

__requires__ = [
    'traits',
    'pyface',
]



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   11/10/2004
#
#------------------------------------------------------------------------------

""" Creates a wxPython specific modal dialog user interface that runs as a
    complete application, using information from the specified UI object.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

# Standard library imports.
import os
import sys

# System library imports.
import wx

# ETS imports.
from pyface.util.guisupport import is_event_loop_running_wx, \
    start_event_loop_wx

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

# File to redirect output to. If '', output goes to stdout.
redirect_filename = ''

KEEP_ALIVE_UIS = set()


def on_ui_destroyed(object, name, old, destroyed):
    """ Remove the UI object from KEEP_ALIVE_UIS.
    """
    assert name == 'destroyed'
    if destroyed:
        assert object in KEEP_ALIVE_UIS
        KEEP_ALIVE_UIS.remove(object)
        object.on_trait_change(on_ui_destroyed, 'destroyed', remove=True)


#-------------------------------------------------------------------------
#  Creates a 'stand-alone' wx Application to display a specified traits UI View:
#-------------------------------------------------------------------------

def view_application(context, view, kind, handler, id, scrollable, args):
    """ Creates a stand-alone wx Application to display a specified traits UI
        View.

    Parameters
    ----------
    context : object or dictionary
        A single object or a dictionary of string/object pairs, whose trait
        attributes are to be edited. If not specified, the current object is
        used.
    view : view object
        A View object that defines a user interface for editing trait attribute
        values.
    kind : string
        The type of user interface window to create. See the
        **traitsui.view.kind_trait** trait for values and
        their meanings. If *kind* is unspecified or None, the **kind**
        attribute of the View object is used.
    handler : Handler object
        A handler object used for event handling in the dialog box. If
        None, the default handler for Traits UI is used.
    scrollable : Boolean
        Indicates whether the dialog box should be scrollable. When set to
        True, scroll bars appear on the dialog box if it is not large enough
        to display all of the items in the view at one time.
    """
    if (kind == 'panel') or ((kind is None) and (view.kind == 'panel')):
        kind = 'modal'

    app = wx.GetApp()
    if app is None or not is_event_loop_running_wx(app):
        return ViewApplication(context, view, kind, handler, id,
                               scrollable, args).ui.result

    ui = view.ui(context,
                 kind=kind,
                 handler=handler,
                 id=id,
                 scrollable=scrollable,
                 args=args)

    # If the UI has not been closed yet, we need to keep a reference to
    # it until it does close.
    if not ui.destroyed:
        KEEP_ALIVE_UIS.add(ui)
        ui.on_trait_change(on_ui_destroyed, 'destroyed')
    return ui.result

#-------------------------------------------------------------------------
#  'ViewApplication' class:
#-------------------------------------------------------------------------


class ViewApplication(wx.App):
    """ Modal window that contains a stand-alone application.
    """
    #-------------------------------------------------------------------------
    #  Initializes the object:
    #-------------------------------------------------------------------------

    def __init__(self, context, view, kind, handler, id, scrollable, args):
        """ Initializes the object.
        """
        self.context = context
        self.view = view
        self.kind = kind
        self.handler = handler
        self.id = id
        self.scrollable = scrollable
        self.args = args

        if os.environ.get('ENABLE_FBI') is not None:
            try:
                from etsdevtools.developer.helper.fbi import enable_fbi
                enable_fbi()
            except:
                pass

        if redirect_filename.strip() != '':
            super(ViewApplication, self).__init__(1, redirect_filename)
        else:
            super(ViewApplication, self).__init__(0)

        # Start the event loop in an IPython-conforming manner.
        start_event_loop_wx(self)

    #-------------------------------------------------------------------------
    #  Handles application initialization:
    #-------------------------------------------------------------------------

    def OnInit(self):
        """ Handles application initialization.
        """
        self.ui = self.view.ui(self.context,
                               kind=self.kind,
                               handler=self.handler,
                               id=self.id,
                               scrollable=self.scrollable,
                               args=self.args)
        return True



def get_all_subclasses(parent, subclassof=None):
    """
    Recursive call to get all classes that have a specified class
    in their ancestry.  The call to __subclasses__ only finds the
    direct, child subclasses of an object, so to find
    grandchildren and objects further down the tree, we have to go
    recursively down each subclasses hierarchy to see if the
    subclasses are of the type we want.

    @param parent: class used to find subclasses
    @type parent: class
    @param subclassof: class used to verify type during recursive calls
    @type subclassof: class
    @returns: list of classes
    """
    if subclassof is None:
        subclassof=parent
    subclasses={}

    # this call only returns immediate (child) subclasses, not
    # grandchild subclasses where there is an intermediate class
    # between the two.
    classes=parent.__subclasses__()
    for kls in classes:
        # FIXME: this seems to return multiple copies of the same class,
        # but I probably just don't understand enough about how python
        # creates subclasses
        # dprint("%s id=%s" % (kls, id(kls)))
        if issubclass(kls,subclassof):
            subclasses[kls] = 1
        # for each subclass, recurse through its subclasses to
        # make sure we're not missing any descendants.
        subs=get_all_subclasses(parent=kls)
        if len(subs)>0:
            for kls in subs:
                subclasses[kls] = 1
    return subclasses.keys()



""" The default workbench action set. """


# Enthought library imports.
from envisage.ui.action.api import Action, ActionSet, Menu


# This module's package.
PKG = '.'.join(__name__.split('.')[:-1])


class DefaultActionSet(ActionSet):
    """ The default workbench action set. """

    menus = [
        Menu(
            name='&File', path='MenuBar',
            groups=['OpenGroup', 'SaveGroup', 'ImportGroup', 'ExitGroup']
        ),

        Menu(
            path='MenuBar',
            class_name='pyface.workbench.action.api:ViewMenuManager'
        ),

        Menu(
            name='&Tools', path='MenuBar',
            groups=['PreferencesGroup']
        ),

        Menu(
            name='&Help', path='MenuBar',
            groups=['AboutGroup']
        )
    ]

    actions = [
        Action(
            path='MenuBar/File', group='ExitGroup',
            class_name=PKG + '.action.api:ExitAction'
        ),

        Action(
            path='MenuBar/Tools', group='PreferencesGroup',
            class_name=PKG + '.action.api:EditPreferencesAction'
        ),

        Action(
            path='MenuBar/Help', group='AboutGroup',
            class_name=PKG + '.action.api:AboutAction'
        ),
    ]

#### EOF ######################################################################



""" An event fired when an extension point's extensions have changed. """


# Enthought library imports.
from traits.api import TraitListEvent


class ExtensionPointChangedEvent(TraitListEvent):
    """ An event fired when an extension point's extensions have changed. """

    def __init__ (self, extension_point_id=None, **kw):
        """ Constructor. """

        # The base class has the 'index', 'removed' and 'added' attributes.
        super(ExtensionPointChangedEvent, self).__init__(**kw)

        # We add the extension point Id.
        self.extension_point_id = extension_point_id

        return

#### EOF ######################################################################



from traits.api import Interface, List, Str


class ISelection(Interface):
    """ Collection of selected items. """

    #: ID of the selection provider that created this selection object.
    provider_id = Str

    def is_empty(self):
        """ Is the selection empty? """


class IListSelection(ISelection):
    """ Selection for ordered sequences of items. """

    #: Selected objects.
    items = List

    #: Indices of the selected objects in the selection provider.
    indices = List



import OpenGL.GL as gl

from glutils import ViewportAs

import render_target

from svg_parser_utils import *
from svg_constants import *
from svg_path import SVGRenderableElement


class SVGPattern(SVGRenderableElement):
    def __init__(self, svg, element, parent):
        SVGRenderableElement.__init__(self, svg, element, parent)

        self.svg = svg
        self.units = element.get('patternContentUnits', 'objectBoundingBox')
        self.x = parse_float(element.get('x', '0.0'))
        self.y = parse_float(element.get('y', '0.0'))
        self.width = parse_float(element.get('width', '1.0'))
        self.height = parse_float(element.get('height', '1.0'))
        self.render_texture = None
        self.render_texture = render_target.RenderTarget(PATTERN_TEX_SIZE, PATTERN_TEX_SIZE)

    def bind_texture(self):
        if not self.render_texture:
            return
        self.render_texture.texture.bind()

    def unbind_texture(self):
        if not self.render_texture:
            return
        self.render_texture.texture.unbind()

    def extents(self):
        min_x, min_y, max_x, max_y = 0, 0, 1, 1

        for p in self.children:
            x0, y0, x1, y1 = p.bounding_box()
            if x0 < min_x: min_x = x0
            if y0 < min_y: min_y = y0
            if x1 > max_x: max_x = x1
            if y1 > max_y: max_y = y1

        return min_x, min_y, max_x, max_y

    def render(self):
        #setup projection matrix..
        min_x, min_y, max_x, max_y = self.extents()

        print "extents"
        print self.extents()

        with self.render_texture:
            with ViewportAs(min_x * self.x, min_y * self.y, max_x * self.width, max_y * self.height, PATTERN_TEX_SIZE,
                            PATTERN_TEX_SIZE):
                gl.glClearColor(0.0, 0.5, 1.0, 1.0)
                gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
                for c in self.children:
                    c.render()




""" Managers for submenus of the Help menu.

    :Copyright: 2008, Enthought Inc.
    :License: BSD
    :Author: Janet Swisher
"""
# This software is provided without warranty under the terms of the BSD
# license included in AppTools/trunk/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!


# Standard library imports.
import logging

# Enthought library imports.
from envisage.api import IExtensionPointUser, IExtensionRegistry
from pyface.action.api import Group, MenuManager
from traits.api import Any, provides, Instance, List, Property

# Local imports.
from action.doc_action import DocAction
from action.demo_action import DemoAction
from action.example_action import ExampleAction
from action.load_url_action import LoadURLAction
from examples_preferences import ExamplesPreferences
from i_help_doc import IHelpDoc
from i_help_code import IHelpCode

# Logging.
logger = logging.getLogger(__name__)

# This module's package.
PKG = '.'.join(__name__.split('.')[:-1])

@provides(IExtensionPointUser)
class HelpSubmenuManager(MenuManager):
    """ Base class for managers of submenus of the Help menu.

        This class is adapted from
        pyface.ui.workbench.action.view_menu_manager.ViewMenuManager.
    """


    ### IExtensionPointUser interface
    extension_registry = Property(Instance(IExtensionRegistry))

    def _get_extension_registry(self):
        return self.window.application.extension_registry

    #### 'ActionManager' interface ############################################

    # All of the groups in the manager.
    groups = List(Group)

    #### 'Private' interface ##################################################

    # The group containing the actual actions.
    _item_group = Any

    ###########################################################################
    # 'ActionManager' interface.
    ###########################################################################

    def _groups_default(self):
        """ Trait initializer. """

        groups = []

        # Add a group containing items for all contributed documents.
        self._item_group = self._create_item_group(self.window)
        groups.append(self._item_group)

        return groups

    ###########################################################################
    # Private interface.
    ###########################################################################
    def _clear_group(self, group):
        """ Remove all items in a group. """

        group.destroy()
        group.clear()

        return

    def _create_item_group(self, window):
        """ Creates a group containing the items. """

        group = Group()
        self._initialize_item_group(window, group)

        return group

    def _initialize_item_group(self, window, group):
        """ Initializes a group containing the items. """
        raise NotImplementedError

class DocumentsMenuManager(HelpSubmenuManager):
    """ Controls the 'Help/Documents' menu.
    """

    #### 'ActionManager' interface ############################################

    # The manager's unique identifier (if it has one).
    id = 'Documents'

    #### 'MenuManager' interface ##############################################

    # The menu manager's name (if the manager is a sub-menu, this is what its
    # label will be).
    name = u'&Documents'

    #### 'DocMenuManager' interface ##########################################

    # The HelpDocs for which this manager displays menu items.
    help_doc_list = List(IHelpDoc, allow_none=True)

    def _help_doc_list_default(self):
        return self.extension_registry.get_extensions(PKG + '.help_docs')

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _initialize_item_group(self, window, group):
        """ Initializes a group containing the items. """

        docs = self.help_doc_list
        docs.sort(None, lambda doc: doc.label)

        for doc in docs:
            #logger.info('Adding Helpaction for "%s", %s' % (doc.label, str(doc)))
            group.append(
                DocAction(name=doc.label, window=window)
                )

        return

class DemosMenuManager(HelpSubmenuManager):
    """ Controls the 'Help/Demos' menu.
    """

    #### 'ActionManager' interface ############################################

    # The manager's unique identifier (if it has one).
    id = 'Demos'

    #### 'MenuManager' interface ##############################################

    # The menu manager's name (if the manager is a sub-menu, this is what its
    # label will be).
    name = u'D&emos'

    #### 'DemoMenuManager' interface ##########################################

    # The HelpCode for which this manager displays menu items.
    help_demo_list = List(IHelpCode, allow_none=True)

    def _help_demo_list_default(self):
        return self.extension_registry.get_extensions(PKG + '.help_demos')

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _initialize_item_group(self, window, group):
        """ Initializes a group containing the items. """

        demos = self.help_demo_list
        demos.sort(None, lambda demo: demo.label)

        for demo in demos:
            group.append(
                DemoAction(name=demo.label, window=window)
                )

        return

class ExamplesMenuManager(HelpSubmenuManager):
    """ Controls the 'Help/Examples' menu.
    """

    #### 'ActionManager' interface ############################################

    # The manager's unique identifier (if it has one).
    id = 'Examples'

    #### 'MenuManager' interface ##############################################

    # The menu manager's name (if the manager is a sub-menu, this is what its
    # label will be).
    name = u'&Examples'

    #### 'ExampleMenuManager' interface ##########################################

    # The HelpCode for which this manager displays menu items.
    help_example_list = List(IHelpCode, allow_none=True)

    def _help_example_list_default(self):
        return self.extension_registry.get_extensions(PKG + '.help_examples')

    # Preferences for examples
    preferences = Instance(ExamplesPreferences, ())

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _initialize_item_group(self, window, group):
        """ Initializes a group containing the items. """

        examples = self.help_example_list
        examples.sort(None, lambda example: example.label)

        for ex in examples:
            group.append(
                ExampleAction(name=ex.label, window=window,
                                  preferences=self.preferences)
                )

        return


class DownloadsMenuManager(HelpSubmenuManager):
    """ Controls the 'Help/Downloads' or 'Downloads'menu.
    """

    #### 'ActionManager' interface ############################################

    # The manager's unique identifier (if it has one).
    id = 'Downloads'

    #### 'MenuManager' interface ##############################################

    # The menu manager's name (if the manager is a sub-menu, this is what its
    # label will be).
    name = u'&Downloads'

    #### 'DocMenuManager' interface ##########################################

    # The URLs for which this manager displays menu items.
    help_download_list = List(IHelpDoc, allow_none=True)

    def _help_download_list_default(self):
        return self.extension_registry.get_extensions(PKG + '.help_downloads')

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _initialize_item_group(self, window, group):
        """ Initializes a group containing the items. """

        urls = self.help_download_list
        #docs.sort(None, lambda doc: doc.label)

        for url in urls:
            group.append(
                LoadURLAction(name=url.label, window=window)
                )

        return
#### EOF ######################################################################



#!/usr/bin/env python

"""
code to read the netcdf unstructured grid standard:

https://github.com/ugrid-conventions/ugrid-conventions/

This code is called by the UGrid class to load into a UGRID object.

    NOTE: passing the UGrid object in to avoid circular references,
    while keeping the netcdf reading code in its own file.

"""

from __future__ import (absolute_import, division, print_function)

import numpy as np
import netCDF4

from .uvar import UVar


def find_mesh_names(nc):
    """
    find all the meshes in an open netcCDF4.DataSet

    :param nc: the netCDF4 Dataset object to look for mesh names in

    NOTE: checks for 2-d topology_dimension
    """
    mesh_names = []
    for varname in nc.variables.keys():
        if is_valid_mesh(nc, varname):
                    mesh_names.append(varname)
    return mesh_names


def is_valid_mesh(nc, varname):
    """
    determine if the given variable name is a valid mesh definition

    :param nc: a netCDF4 Dataset to check

    :param varname: name of the candidate mesh variable

    """
    try:
        mesh_var = nc.variables[varname]
    except KeyError:
        return False
    try:
        if (mesh_var.cf_role.strip() == 'mesh_topology' and
           int(mesh_var.topology_dimension) == 2):
            return True
    except AttributeError:
            # not a valid mesh variable
        return False

# Defining properties of various connectivity arrays
# so that the same code can load all of them.
grid_defs = [{'grid_attr': 'faces',  # Name in UGrid object.
              'role': 'face_node_connectivity',  # Name in mesh variable.
              'num_ind': 3,  # Number of idx (3 for faces, 2 for segments).
              },
             {'grid_attr': 'face_face_connectivity',  # Name in UGrid object.
              'role': 'face_face_connectivity',  # Name in mesh variable.
              'num_ind': 3,  # Number of idx (3 for faces, 2 for segments).
              },
             {'grid_attr': 'boundaries',  # Name in UGrid object.
              'role': 'boundary_node_connectivity',  # Name in mesh var.
              'num_ind': 2,  # Number of idx (3 for faces, 2 for segments).
              },
             {'grid_attr': 'edges',  # Name in UGrid object.
              'role': 'edge_node_connectivity',  # Name in mesh variable.
              'num_ind': 2,  # Number of idx (3 for faces, 2 for segments).
              },
             ]
# definitions for various coordinate arrays
coord_defs = [{'grid_attr': 'nodes',  # Attribute name in UGrid object.
               'role': 'node_coordinates',  # Attribute name in mesh variable.
               'required': True,  # Is this required?
               },
              {'grid_attr': 'face_coordinates',  # Name in UGrid object.
               'role': 'face_coordinates',  # Name in mesh variable.
               'required': False,  # Is this required?
               },
              {'grid_attr': 'edge_coordinates',  # Name in UGrid object.
               'role': 'edge_coordinates',  # Name in mesh variable.
               'required': False,  # is this required?
               },
              {'grid_attr': 'boundary_coordinates',  # Name in UGrid object.
               'role': 'boundary_coordinates',  # Name in mesh variable.
               'required': False,  # Is this required?
               },
              ]


def load_grid_from_nc_dataset(nc, grid, mesh_name=None, load_data=True):
    """
    loads UGrid object from a netCDF4.DataSet object, adding the data
    to the passed-in grid object.

    It will load the mesh specified, or look
    for the first one it finds if none is specified

    :param nc: netcdf Dataset to be loaded up
    :type nc: netCDF4 Dataset object

    :param grid: the grid object to put the mesh and data into.
    :type grid: UGrid object.

    :param mesh_name=None: name of the mesh to load
    :type mesh_name: string

    :param load_data=False: flag to indicate whether you want to load the
                            associated data or not.  The mesh will be loaded
                            in any case.  If False, only the mesh will be
                            loaded.  If True, then all the data associated
                            with the mesh will be loaded.  This could be huge!
    :type load_data: boolean

    NOTE: passing the UGrid object in to avoid circular references,
    while keeping the netcdf reading code in its own file.
    """
    ncvars = nc.variables

    # Get the mesh_name.
    if mesh_name is None:
        # Find the mesh.
        meshes = find_mesh_names(nc)
        if len(meshes) == 0:
            msg = "There are no standard-conforming meshes in {}".format
            raise ValueError(msg(nc.filepath))
        if len(meshes) > 1:
            msg = "There is more than one mesh in the file: {!r}".format
            raise ValueError(msg(meshes))
        mesh_name = meshes[0]
    else:
        if not is_valid_mesh(nc, mesh_name):
            msg = "Mesh: {} is not in {}".format
            raise ValueError(msg(mesh_name, nc.filepath))

    grid.mesh_name = mesh_name
    mesh_var = ncvars[mesh_name]

    # Load the coordinate variables.
    for defs in coord_defs:
        try:
            coord_names = mesh_var.getncattr(defs['role']).strip().split()
            coord_vars = [nc.variables[name] for name in coord_names]
        except AttributeError:
            if defs['required']:
                msg = "Mesh variable must include {} attribute.".format
                raise ValueError(msg(defs['role']))
            continue
        except KeyError:
            msg = ("File must include {} variables for {} "
                   "named in mesh variable.").format
            raise ValueError(msg(coord_names, defs['role']))

        coord_vars = [nc.variables[name] for name in coord_names]
        num_node = len(coord_vars[0])
        nodes = np.empty((num_node, 2), dtype=np.float64)
        for var in coord_vars:
            try:
                standard_name = var.standard_name
            except AttributeError:
                # CF does not require a standard name, look in units, instead.
                try:
                    units = var.units
                except AttributeError:
                    msg = ("The {} variable doesn't contain units "
                           "attribute: required by CF").format
                    raise ValueError(msg(var))
                # CF accepted units attributes for longitude.
                if units in ('degrees_east', 'degree_east', 'degree_E',
                             'degrees_E', 'degreeE', 'degreesE'):
                        standard_name = 'longitude'
                # CF accepted units attributes for longitude.
                elif units in ('degrees_north', 'degree_north', 'degree_N',
                               'degrees_N', 'degreeN', 'degreesN'):
                        standard_name = 'latitude'
                else:
                    msg = ("{} variable's units value ({}) doesn't look "
                           "like latitude or longitude").format
                    raise ValueError(msg(var, units))
            if standard_name == 'latitude':
                nodes[:, 1] = var[:]
            elif standard_name == 'longitude':
                nodes[:, 0] = var[:]
            else:
                raise ValueError('Node coordinates standard_name is neither '
                                 '"longitude" nor "latitude" ')
        setattr(grid, defs['grid_attr'], nodes)

    # Load assorted connectivity arrays.
    for defs in grid_defs:
        try:
            try:
                var = nc.variables[mesh_var.getncattr(defs['role'])]
            except AttributeError:  # This connectivity array isn't there.
                continue
            array = var[:, :]
            # Fortran order, instead of C order, transpose the array
            # logic below will fail for 3 node or two edge grids.
            if array.shape[0] == defs['num_ind']:
                array = array.T
            try:
                start_index = int(var.start_index)
            except AttributeError:
                start_index = 0
            if start_index >= 1:
                array -= start_index
                # Check for flag value.
                try:
                    # FIXME: This won't work for more than one flag value.
                    flag_value = var.flag_values
                    array[array == flag_value-start_index] = flag_value
                except AttributeError:
                    pass
            setattr(grid, defs['grid_attr'], array)
        except KeyError:
            pass  # OK not to have this...

    # Load the associated data:

    if load_data:
        # Look for data arrays -- they should have a "location" attribute.
        for name, var in nc.variables.items():
            # Data Arrays should have "location" and "mesh" attributes.
            try:
                location = var.location
                # The mesh attribute should match the mesh we're loading:
                if var.mesh != mesh_name:
                    continue
            except AttributeError:
                continue

            # Get the attributes.
            # FIXME: Is there a way to get the attributes a Variable directly?
            attributes = {n: var.getncattr(n) for n in var.ncattrs()
                          if n not in ('location', 'coordinates', 'mesh')}

            # Trick with the name: FIXME: Is this a good idea?
            name = name.lstrip(mesh_name).lstrip('_')
            uvar = UVar(name, data=var[:],
                        location=location, attributes=attributes)
            grid.add_data(uvar)


def load_grid_from_ncfilename(filename, grid, mesh_name=None, load_data=True):
    """
    loads UGrid object from a netcdf file, adding the data
    to the passed-in grid object.

    It will load the mesh specified, or look
    for the first one it finds if none is specified

    :param filename: filename or OpenDAP url of dataset.

    :param grid: the grid object to put the mesh and data into.
    :type grid: UGrid object.

    :param mesh_name=None: name of the mesh to load
    :type mesh_name: string

    :param load_data=False: flag to indicate whether you want to load the
                            associated data or not.  The mesh will be loaded
                            in any case.  If False, only the mesh will be
                            loaded.  If True, then all the data associated
                            with the mesh will be loaded.  This could be huge!
    :type load_data: boolean
    """

    with netCDF4.Dataset(filename, 'r') as nc:
        load_grid_from_nc_dataset(nc, grid, mesh_name, load_data)



#!/usr/bin/env python
"""
fs.browsewin
============

Creates a window which can be used to browse the contents of a filesystem.
To use, call the 'browse' method with a filesystem object. Double click a file
or directory to display its properties.

Requires wxPython.

"""

import wx
import wx.gizmos

from fs.path import isdotfile, pathsplit
from fs.errors import FSError

class InfoFrame(wx.Frame):

    def __init__(self, parent, path, desc, info):
        wx.Frame.__init__(self, parent, -1, style=wx.DEFAULT_FRAME_STYLE, size=(500, 500))

        self.SetTitle("FS Object info - %s (%s)" % (path, desc))

        keys = info.keys()
        keys.sort()

        self.list_ctrl = wx.ListCtrl(self, -1, style=wx.LC_REPORT|wx.SUNKEN_BORDER)

        self.list_ctrl.InsertColumn(0, "Key")
        self.list_ctrl.InsertColumn(1, "Value")

        self.list_ctrl.SetColumnWidth(0, 190)
        self.list_ctrl.SetColumnWidth(1, 300)

        for key in sorted(keys, key=lambda k:k.lower()):
            self.list_ctrl.Append((key, unicode(info.get(key))))
            
        self.Center()



class BrowseFrame(wx.Frame):

    def __init__(self, fs, hide_dotfiles=False):

        wx.Frame.__init__(self, None, size=(1000, 600))

        self.fs = fs
        self.hide_dotfiles = hide_dotfiles
        self.SetTitle("FS Browser - " + unicode(fs))

        self.tree = wx.gizmos.TreeListCtrl(self, -1, style=wx.TR_DEFAULT_STYLE | wx.TR_HIDE_ROOT)

        self.tree.AddColumn("File System")
        self.tree.AddColumn("Description")
        self.tree.AddColumn("Size")
        self.tree.AddColumn("Created")

        self.tree.SetColumnWidth(0, 300)
        self.tree.SetColumnWidth(1, 250)
        self.tree.SetColumnWidth(2, 150)
        self.tree.SetColumnWidth(3, 250)

        self.root_id = self.tree.AddRoot('root', data=wx.TreeItemData( {'path':"/", 'expanded':False} ))

        rid = self.tree.GetItemData(self.root_id)

        isz = (16, 16)
        il = wx.ImageList(isz[0], isz[1])
        self.fldridx     = il.Add(wx.ArtProvider_GetBitmap(wx.ART_FOLDER,      wx.ART_OTHER, isz))
        self.fldropenidx = il.Add(wx.ArtProvider_GetBitmap(wx.ART_FILE_OPEN,   wx.ART_OTHER, isz))
        self.fileidx     = il.Add(wx.ArtProvider_GetBitmap(wx.ART_NORMAL_FILE, wx.ART_OTHER, isz))

        self.tree.SetImageList(il)
        self.il = il

        self.tree.SetItemImage(self.root_id, self.fldridx, wx.TreeItemIcon_Normal)
        self.tree.SetItemImage(self.root_id, self.fldropenidx, wx.TreeItemIcon_Expanded)

        self.Bind(wx.EVT_TREE_ITEM_EXPANDING, self.OnItemExpanding)
        self.Bind(wx.EVT_TREE_ITEM_ACTIVATED, self.OnItemActivated)

        wx.CallAfter(self.OnInit)

    def OnInit(self):

        self.expand(self.root_id)


    def expand(self, item_id):

        item_data = self.tree.GetItemData(item_id).GetData()

        path = item_data["path"]

        if not self.fs.isdir(path):
            return

        if item_data['expanded']:
            return

        try:
            paths = ( [(True, p) for p in self.fs.listdir(path, absolute=True, dirs_only=True)] +
                      [(False, p) for p in self.fs.listdir(path, absolute=True, files_only=True)] )
        except FSError, e:
            msg = "Failed to get directory listing for %s\n\nThe following error was reported:\n\n%s" % (path, e)
            wx.MessageDialog(self, msg, "Error listing directory", wx.OK).ShowModal()
            paths = []
            
                
        #paths = [(self.fs.isdir(p), p) for p in self.fs.listdir(path, absolute=True)]
        
        if self.hide_dotfiles:
            paths = [p for p in paths if not isdotfile(p[1])]

        if not paths:
            #self.tree.SetItemHasChildren(item_id, False)
            #self.tree.Collapse(item_id)
            return

        paths.sort(key=lambda p:(not p[0], p[1].lower()))

        for is_dir, new_path in paths:

            name = pathsplit(new_path)[-1]

            new_item = self.tree.AppendItem(item_id, name, data=wx.TreeItemData({'path':new_path, 'expanded':False}))

            info = self.fs.getinfo(new_path)

            if is_dir:

                self.tree.SetItemHasChildren(new_item)
                self.tree.SetItemImage(new_item, self.fldridx, 0, wx.TreeItemIcon_Normal)
                self.tree.SetItemImage(new_item, self.fldropenidx, 0, wx.TreeItemIcon_Expanded)

                self.tree.SetItemText(new_item, "", 2)

                ct = info.get('created_time', None)
                if ct is not None:
                    self.tree.SetItemText(new_item, ct.ctime(), 3)
                else:
                    self.tree.SetItemText(new_item, 'unknown', 3)

            else:
                self.tree.SetItemImage(new_item, self.fileidx, 0, wx.TreeItemIcon_Normal)

                self.tree.SetItemText(new_item, str(info.get('size', '?'))+ " bytes", 2)

                ct = info.get('created_time', None)
                if ct is not None:
                    self.tree.SetItemText(new_item, ct.ctime(), 3)
                else:
                    self.tree.SetItemText(new_item, 'unknown', 3)

            self.tree.SetItemText(new_item, self.fs.desc(new_path), 1)

        item_data['expanded'] = True
        self.tree.Expand(item_id)


    def OnItemExpanding(self, e):

        self.expand(e.GetItem())
        e.Skip()

    def OnItemActivated(self, e):

        item_data = self.tree.GetItemData(e.GetItem()).GetData()
        path = item_data["path"]
        info = self.fs.getinfo(path)

        info_frame = InfoFrame(self, path, self.fs.desc(path), info)        
        info_frame.Show()
        info_frame.CenterOnParent()


def browse(fs, hide_dotfiles=False):
    """Displays a window containing a tree control that displays an FS
    object. Double-click a file/folder to display extra info.

    :param fs: A filesystem object
    :param hide_dotfiles: If True, files and folders that begin with a dot will be hidden

    """

    app = wx.PySimpleApp()
    frame = BrowseFrame(fs, hide_dotfiles=hide_dotfiles)
    frame.Show()
    app.MainLoop()


if __name__ == "__main__":
    from osfs import OSFS
    home_fs = OSFS("~/")
    browse(home_fs, True)



#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   06/05/2007
#
#-------------------------------------------------------------------------

""" Traits UI 'display only' image editor.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from pyface.ui_traits import Image
from traits.api import Bool, Property

from ..basic_editor_factory import BasicEditorFactory

from ..toolkit import toolkit_object

#-------------------------------------------------------------------------
#  'ImageEditor' editor factory class:
#-------------------------------------------------------------------------


class ImageEditor(BasicEditorFactory):

    # The editor class to be created:
    klass = Property

    # The optional image resource to be displayed by the editor (if not
    # specified, the editor's object value is used as the ImageResource to
    # display):
    image = Image

    # The following traits are currently supported on Qt only

    # Whether or not to scale the image to fit the available space
    scale = Bool

    # Whether or not to scale the image larger than the original when scaling
    allow_upscaling = Bool

    # Whether or not to preserve the aspect ratio when scaling
    preserve_aspect_ratio = Bool

    # Whether or not to allow the image to be clipped when not scaling
    allow_clipping = Bool

    def _get_klass(self):
        """ Returns the editor class to be instantiated.
        """
        return toolkit_object('image_editor:_ImageEditor')

#-- EOF -----------------------------------------------------------------------



import os

from traits.trait_base import get_resource_path

import logging
log = logging.getLogger(__name__)


class MissingDocumentationError(RuntimeError):
    pass


root_resource_path = None

help_dirs = ["omnivore/help", "omnivore8bit/help"]


def get_htmlhelp(name):
    global root_resource_path

    # resource path will point to the omnivore/templates directory
    if root_resource_path is None:
        path = get_resource_path(1)
        root_resource_path = os.path.normpath(os.path.join(path, "../.."))
        log.debug("resource path: %s" % root_resource_path)
    attempts = []
    for subdir in help_dirs:
        pathname = os.path.normpath(os.path.join(root_resource_path, subdir, name))
        log.debug("Checking for htmlhelp at %s" % pathname)
        if os.path.exists(pathname):
            log.debug("Found htmlhelp for %s: %s" % (name, pathname))
            return pathname
        attempts.append(pathname)
    else:
        log.debug("No htmlhelp found for %s in %s" % (name, attempts))
        raise MissingDocumentationError("Unable to locate %s help files; installation error?\nThe files should be in one of the following:\n%s\n\nbut were not found." % (name, "\n  * " + "\n  * ".join([os.path.dirname(d) for d in attempts])))



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/18/2004
#
#------------------------------------------------------------------------------

""" Defines the Include class, which is used to represent a substitutable
    element within a user interface View.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import Str

from .view_element import ViewSubElement

#-------------------------------------------------------------------------
#  'Include' class:
#-------------------------------------------------------------------------


class Include(ViewSubElement):
    """ A substitutable user interface element, i.e., a placeholder in a view
    definition.

    When a view object constructs an attribute-editing window, any Include
    objects within the view definition are replaced with a group or item
    defined elsewhere in the object's inheritance tree, based on matching of the
    name of the element. If no matching element is found, the Include object
    is ignored.

    An Include object can reference a group or item attribute on a parent class
    or on a subclass. For example, the following class contains a view
    definition that provides for the possibility that a subclass might add
    "extra" attributes in the middle of the view::

        class Person(HasTraits):
            name = Str
            age = Int
            person_view = View('name', Include('extra'), 'age', kind='modal')

    If you directly create an instance of Person, and edit its attributes,
    the Include object is ignored.

    The following class extends Person, and defines a group of "extra"
    attributes to add to the view defined on Person::

        class LocatedPerson(Person):
            street = Str
            city = Str
            state = Str
            zip = Int
            extra = Group('street', 'city', 'state', 'zip')

    The attribute-editing window for an instance of LocatedPerson displays
    editors for these extra attributes.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # The name of the substitutable content
    id = Str

    #-------------------------------------------------------------------------
    #  Initializes the object:
    #-------------------------------------------------------------------------

    def __init__(self, id, **traits):
        """ Initializes the Include object.
        """
        super(ViewSubElement, self).__init__(**traits)
        self.id = id

    #-------------------------------------------------------------------------
    #  Returns a 'pretty print' version of the Include:
    #-------------------------------------------------------------------------

    def __repr__(self):
        """ Returns a "pretty print" version of the Include object.
        """
        return "<%s>" % self.id



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/25/2004
#
#------------------------------------------------------------------------------

""" Defines helper functions and classes used to define wxPython-based trait
    editors and trait editor factories.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from operator import itemgetter

import wx
import wx.lib.scrolledpanel

import sys

from os.path \
    import join, dirname, abspath

from traits.api \
    import HasPrivateTraits, Enum, CTrait, Instance, Any, Int, \
    Event, Bool, BaseTraitHandler, TraitError

from traitsui.ui_traits \
    import convert_image, SequenceTypes

from pyface.timer.api \
    import do_later

from constants \
    import standard_bitmap_width, screen_dx, screen_dy

from editor \
    import Editor

#-------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------

# Layout orientation for a control and its associated editor
Orientation = Enum('horizontal', 'vertical')

#-------------------------------------------------------------------------
#  Data:
#-------------------------------------------------------------------------

# Bitmap cache dictionary (indexed by filename)
_bitmap_cache = {}

### NOTE: This needs major improvements:

app_path = None
traits_path = None

#-------------------------------------------------------------------------
#  Convert an image file name to a cached bitmap:
#-------------------------------------------------------------------------


def bitmap_cache(name, standard_size, path=None):
    """ Converts an image file name to a cached bitmap.
    """
    global app_path, traits_path

    if name[:1] == '@':
        image = convert_image(name.replace(' ', '_').lower())
        if image is not None:
            return image.create_image().ConvertToBitmap()

    if path is None:
        if traits_path is None:
            import traitsui.wx
            traits_path = join(dirname(traitsui.wx.__file__),
                               'images')
        path = traits_path
    elif path == '':
        if app_path is None:
            app_path = join(dirname(sys.argv[0]), '..', 'images')
        path = app_path

    filename = abspath(join(path, name.replace(' ', '_').lower() + '.gif'))
    bitmap = _bitmap_cache.get(filename + ('*'[not standard_size:]))
    if bitmap is not None:
        return bitmap

    std_bitmap = bitmap = wx.BitmapFromImage(wx.Image(filename))
    _bitmap_cache[filename] = bitmap

    dx = bitmap.GetWidth()
    if dx < standard_bitmap_width:
        dy = bitmap.GetHeight()
        std_bitmap = wx.EmptyBitmap(standard_bitmap_width, dy)
        dc1 = wx.MemoryDC()
        dc2 = wx.MemoryDC()
        dc1.SelectObject(std_bitmap)
        dc2.SelectObject(bitmap)
        dc1.SetPen(wx.TRANSPARENT_PEN)
        dc1.SetBrush(wx.WHITE_BRUSH)
        dc1.DrawRectangle(0, 0, standard_bitmap_width, dy)
        dc1.Blit((standard_bitmap_width - dx) / 2, 0, dx, dy, dc2, 0, 0)

    _bitmap_cache[filename + '*'] = std_bitmap

    if standard_size:
        return std_bitmap

    return bitmap

#-------------------------------------------------------------------------
#  Returns an appropriate width for a wxChoice widget based upon the list of
#  values it contains:
#-------------------------------------------------------------------------


def choice_width(values):
    """ Returns an appropriate width for a wxChoice widget based upon the list
        of values it contains:
    """
    return max([len(x) for x in values]) * 6

#-------------------------------------------------------------------------
#  Saves the user preference items for a specified UI:
#-------------------------------------------------------------------------


def save_window(ui):
    """ Saves the user preference items for a specified UI.
    """
    control = ui.control
    ui.save_prefs(control.GetPositionTuple() + control.GetSizeTuple())

#-------------------------------------------------------------------------
#  Restores the user preference items for a specified UI:
#-------------------------------------------------------------------------


def restore_window(ui, is_popup=False):
    """ Restores the user preference items for a specified UI.
    """
    prefs = ui.restore_prefs()
    if prefs is not None:
        x, y, dx, dy = prefs

        # Check to see if the window's position is within a display.
        # If it is not entirely within 1 display, move it and/or
        # resize it to the closest window

        closest = find_closest_display(x, y)
        x, y, dx, dy = get_position_for_display(x, y, dx, dy, closest)

        if is_popup:
            position_window(ui.control, dx, dy)
        else:
            if (dx, dy) == (0, 0):
                # The window was saved minimized
                ui.control.SetDimensions(x, y, -1, -1)
            else:
                ui.control.SetDimensions(x, y, dx, dy)


def find_closest_display(x, y):
    """ For a virtual screen position, find the closest display.
        There are a few reasons to use this function:
            * the number of displays changed
            * the size of the displays changed
            * the orientation of one or more displays changed.
    """
    closest = None
    for display_num in range(wx.Display.GetCount()):
        display = wx.Display(display_num)
        if closest is None:
            closest = display
        else:
            def _distance(x, y, display):
                dis_x, dis_y, dis_w, dis_h = display.GetGeometry()
                dis_mid_x = (dis_x + dis_w) / 2
                dis_mid_y = (dis_y + dis_h) / 2

                return (x - dis_mid_x)**2 + (y - dis_mid_y)**2

            if _distance(x, y, display) < _distance(x, y, closest):
                closest = display

    return closest


def get_position_for_display(x, y, dx, dy, display):
    """ calculates a valid position and size for a window to fit
        inside a display
    """
    dis_x, dis_y, dis_w, dis_h = display.GetGeometry()
    dx = min(dx, dis_w)
    dy = min(dy, dis_h)
    if ((x + dx) > (dis_x + dis_w)) or (x < dis_x):
        x = dis_x
    if ((y + dy) > (dis_y + dis_h)) or (y < dis_y):
        y = dis_y

    return x, y, dx, dy

#-------------------------------------------------------------------------
#  Positions a window on the screen with a specified width and height so that
#  the window completely fits on the screen if possible:
#-------------------------------------------------------------------------


def position_window(window, width=None, height=None, parent=None):
    """ Positions a window on the screen with a specified width and height so
        that the window completely fits on the screen if possible.
    """
    dx, dy = window.GetSizeTuple()
    width = width or dx
    height = height or dy

    if parent is None:
        parent = window._parent

    if parent is None:
        # Center the popup on the screen:
        window.SetDimensions((screen_dx - width) / 2,
                             (screen_dy - height) / 2, width, height)
        return

    # Calculate the desired size of the popup control:
    if isinstance(parent, wx.Window):
        x, y = parent.ClientToScreenXY(0, 0)
        parent_dx, parent_dy = parent.GetSizeTuple()
    else:
        # Special case of parent being a screen position and size tuple (used
        # to pop-up a dialog for a table cell):
        x, y, parent_dx, parent_dy = parent

    adjacent = (getattr(window, '_kind', 'popup') == 'popup')
    width = min(max(parent_dx, width), screen_dx)
    height = min(height, screen_dy)

    closest = find_closest_display(x, y)

    if adjacent:
        y += parent_dy

    x, y, dx, dy = get_position_for_display(x, y, width, height, closest)

    window.SetDimensions(x, y, dx, dy)

#-------------------------------------------------------------------------
#  Returns the top-level window for a specified control:
#-------------------------------------------------------------------------


def top_level_window_for(control):
    """ Returns the top-level window for a specified control.
    """
    parent = control.GetParent()
    while parent is not None:
        control = parent
        parent = control.GetParent()

    return control

#-------------------------------------------------------------------------
#  Recomputes the mappings for a new set of enumeration values:
#-------------------------------------------------------------------------


def enum_values_changed(values):
    """ Recomputes the mappings for a new set of enumeration values.
    """

    if isinstance(values, dict):
        data = [(unicode(v), n) for n, v in values.items()]
        if len(data) > 0:
            data.sort(key=itemgetter(0))
            col = data[0][0].find(':') + 1
            if col > 0:
                data = [(n[col:], v) for n, v in data]
    elif not isinstance(values, SequenceTypes):
        handler = values
        if isinstance(handler, CTrait):
            handler = handler.handler
        if not isinstance(handler, BaseTraitHandler):
            raise TraitError("Invalid value for 'values' specified")
        if handler.is_mapped:
            data = [(unicode(n), n) for n in handler.map.keys()]
            data.sort(key=itemgetter(0))
        else:
            data = [(unicode(v), v) for v in handler.values]
    else:
        data = [(unicode(v), v) for v in values]

    names = [x[0] for x in data]
    mapping = {}
    inverse_mapping = {}
    for name, value in data:
        mapping[name] = value
        inverse_mapping[value] = name

    return (names, mapping, inverse_mapping)

#-------------------------------------------------------------------------
#  Disconnects a wx event handle from its associated control:
#-------------------------------------------------------------------------


def disconnect(control, *events):
    """ Disconnects a wx event handle from its associated control.
    """
    id = control.GetId()
    for event in events:
        event(control, id, None)


def disconnect_no_id(control, *events):
    """ Disconnects a wx event handle from its associated control.
    """
    for event in events:
        event(control, None)

#-------------------------------------------------------------------------
#  Creates a wx.Panel that correctly sets its background color to be the same
#  as its parents:
#-------------------------------------------------------------------------


class TraitsUIPanel(wx.Panel):

    def __init__(self, parent, *args, **kw):
        """ Creates a wx.Panel that correctly sets its background color to be
            the same as its parents.
        """
        bg_color = kw.pop('bg_color', None)
        wx.Panel.__init__(self, parent, *args, **kw)

        wx.EVT_CHILD_FOCUS(self, self.OnChildFocus)

        if bg_color:
            self.SetBackgroundColour(bg_color)
        else:
            # Mac/Win needs this, otherwise background color is black
            attr = self.GetDefaultAttributes()
            self.SetBackgroundColour(attr.colBg)

    def OnChildFocus(self, event):
        """ If the ChildFocusEvent contains one of the Panel's direct children,
            then we will Skip it to let it pass up the widget hierarchy.

            Otherwise, we consume the event to make sure it doesn't go any
            farther. This works around a problem in wx 2.8.8.1 where each Panel
            in a nested hierarchy generates many events that might consume too
            many resources. We do, however, let one event bubble up to the top
            so that it may inform a top-level ScrolledPanel that a descendant
            has acquired focus.
        """
        if event.GetWindow() in self.GetChildren():
            event.Skip()

#-------------------------------------------------------------------------
#  'ChildFocusOverride' class:
#-------------------------------------------------------------------------

# PyEvtHandler was only introduced in wxPython 2.8.8. Fortunately, it is only
# necessary in wxPython 2.8.8.
if wx.__version__ < '2.8.8':

    class ChildFocusOverride(object):

        def __init__(self, window):
            # Set up the event listener.
            window.Bind(wx.EVT_CHILD_FOCUS, window.OnChildFocus)

else:

    class ChildFocusOverride(wx.PyEvtHandler):
        """ Override the scroll-to-focus behaviour in wx 2.8.8's ScrolledWindow
            C++ implementation for ScrolledPanel.

            Instantiating this class with the ScrolledPanel will register the
            new instance as the event handler for the panel.
        """

        def __init__(self, window):
            self.window = window
            wx.PyEvtHandler.__init__(self)

            # Make self the event handler for the window.
            window.PushEventHandler(self)

        def ProcessEvent(self, event):
            if isinstance(event, wx.ChildFocusEvent):
                # Handle this one with our code and don't let the C++ event handler
                # get it.
                return self.window.OnChildFocus(event)
            else:
                # Otherwise, just pass this along in the event handler chain.
                result = self.GetNextHandler().ProcessEvent(event)
                return result

#-------------------------------------------------------------------------
#  'TraitsUIScrolledPanel' class:
#-------------------------------------------------------------------------


class TraitsUIScrolledPanel(wx.lib.scrolledpanel.ScrolledPanel):

    def __init__(self, parent, id=-1, pos=wx.DefaultPosition,
                 size=wx.DefaultSize, style=wx.TAB_TRAVERSAL,
                 name="scrolledpanel"):

        wx.PyScrolledWindow.__init__(self, parent, id, pos=pos, size=size,
                                     style=style, name=name)
        # FIXME: The ScrolledPanel class calls SetInitialSize in its __init__
        # method, but for some reason, that leads to very a small window size.
        # Calling SetSize seems to work okay, but its not clear why
        # SetInitialSize does not work.
        self.SetSize(size)
        self.SetBackgroundColour(parent.GetBackgroundColour())

        # Override the C++ ChildFocus event handler:
        ChildFocusOverride(self)

    def OnChildFocus(self, event):
        """ Handle a ChildFocusEvent.

        Returns a boolean so it can be used as a library call, too.
        """
        self.ScrollChildIntoView(self.FindFocus())

        return True

    def ScrollChildIntoView(self, child):
        """ Scrolls the panel such that the specified child window is in view.
            This method overrides the original in the base class so that
            nested subpanels are handled correctly.
        """
        if child is None:
            return

        sppux, sppuy = self.GetScrollPixelsPerUnit()
        vsx, vsy = self.GetViewStart()

        crx, cry, crdx, crdy = child.GetRect()
        subwindow = child.GetParent()
        while subwindow not in [self, None]:
            # Make sure that the descendant's position information is relative
            # to us, not its local parent.
            pwx, pwy = subwindow.GetRect()[:2]
            crx, cry = crx + pwx, cry + pwy
            subwindow = subwindow.GetParent()

        cr = wx.Rect(crx, cry, crdx, crdy)

        client_size = self.GetClientSize()
        new_vsx, new_vsy = -1, -1

        # Is it before the left edge?
        if (cr.x < 0) and (sppux > 0):
            new_vsx = vsx + (cr.x / sppux)

        # Is it above the top?
        if (cr.y < 0) and (sppuy > 0):
            new_vsy = vsy + (cr.y / sppuy)

        # For the right and bottom edges, scroll enough to show the whole
        # control if possible, but if not just scroll such that the top/left
        # edges are still visible:

        # Is it past the right edge ?
        if (cr.right > client_size.width) and (sppux > 0):
            diff = (cr.right - client_size.width) / sppux
            if (cr.x - (diff * sppux)) > 0:
                new_vsx = vsx + diff + 1
            else:
                new_vsx = vsx + (cr.x / sppux)

        # Is it below the bottom ?
        if (cr.bottom > client_size.height) and (sppuy > 0):
            diff = (cr.bottom - client_size.height) / sppuy
            if (cr.y - (diff * sppuy)) > 0:
                new_vsy = vsy + diff + 1
            else:
                new_vsy = vsy + (cr.y / sppuy)

        # Perform the scroll if any adjustments are needed:
        if (new_vsx != -1) or (new_vsy != -1):
            self.Scroll(new_vsx, new_vsy)

#-------------------------------------------------------------------------
#  Initializes standard wx event handlers for a specified control and object:
#-------------------------------------------------------------------------

# Standard wx event handlers:
handlers = (
    (wx.EVT_ERASE_BACKGROUND, '_erase_background'),
    (wx.EVT_PAINT, '_paint'),
    (wx.EVT_SIZE, '_size'),
    (wx.EVT_LEFT_DOWN, '_left_down'),
    (wx.EVT_LEFT_UP, '_left_up'),
    (wx.EVT_LEFT_DCLICK, '_left_dclick'),
    (wx.EVT_MIDDLE_DOWN, '_middle_down'),
    (wx.EVT_MIDDLE_UP, '_middle_up'),
    (wx.EVT_MIDDLE_DCLICK, '_middle_dclick'),
    (wx.EVT_RIGHT_DOWN, '_right_down'),
    (wx.EVT_RIGHT_UP, '_right_up'),
    (wx.EVT_RIGHT_DCLICK, '_right_dclick'),
    (wx.EVT_MOTION, '_motion'),
    (wx.EVT_ENTER_WINDOW, '_enter'),
    (wx.EVT_LEAVE_WINDOW, '_leave'),
    (wx.EVT_MOUSEWHEEL, '_wheel')
)


def init_wx_handlers(control, object, prefix=''):
    """ Initializes a standard set of wx event handlers for a specified control
        and object using a specified prefix.
    """
    global handlers

    for handler, name in handlers:
        method = getattr(object, prefix + name, None)
        if method is not None:
            handler(control, method)

#-------------------------------------------------------------------------
#  Safely tries to pop up an FBI window if etsdevtools.debug is installed
#-------------------------------------------------------------------------


def open_fbi():
    try:
        from etsdevtools.developer.helper.fbi import if_fbi
        if not if_fbi():
            import traceback
            traceback.print_exc()
    except ImportError:
        pass

#-------------------------------------------------------------------------
#  'GroupEditor' class:
#-------------------------------------------------------------------------


class GroupEditor(Editor):

    #-------------------------------------------------------------------------
    #  Initializes the object:
    #-------------------------------------------------------------------------

    def __init__(self, **traits):
        """ Initializes the object.
        """
        self.set(**traits)

#-------------------------------------------------------------------------
#  'PopupControl' class:
#-------------------------------------------------------------------------


class PopupControl(HasPrivateTraits):

    #-- Constructor Traits ---------------------------------------------------

    # The control the popup should be positioned relative to:
    control = Instance(wx.Window)

    # The minimum width of the popup:
    width = Int

    # The minimum height of the popup:
    height = Int

    # Should the popup be resizable?
    resizable = Bool(False)

    #-- Public Traits --------------------------------------------------------

    # The value (if any) set by the popup control:
    value = Any

    # Event fired when the popup control is closed:
    closed = Event

    #-- Private Traits -------------------------------------------------------

    # The popup control:
    popup = Instance(wx.Window)

    #-- Public Methods -------------------------------------------------------

    def __init__(self, **traits):
        """ Initializes the object.
        """
        super(PopupControl, self).__init__(**traits)

        style = wx.SIMPLE_BORDER
        if self.resizable:
            style = wx.RESIZE_BORDER

        self.popup = popup = wx.Frame(None, -1, '', style=style)
        wx.EVT_ACTIVATE(popup, self._on_close_popup)
        self.create_control(popup)
        self._position_control()
        popup.Show()

    def create_control(self):
        """ Creates the control.

            Must be overridden by a subclass.
        """
        raise NotImplementedError

    def dispose(self):
        """ Called when the popup is being closed to allow any custom clean-up.

            Can be overridden by a subclass.
        """
        pass

    #-- Event Handlers -------------------------------------------------------

    def _value_changed(self, value):
        """ Handles the 'value' being changed.
        """
        do_later(self._close_popup)

    #-- Private Methods ------------------------------------------------------

    def _position_control(self):
        """ Initializes the popup control's initial position and size.
        """
        # Calculate the desired size of the popup control:
        px, cy = self.control.ClientToScreenXY(0, 0)
        cdx, cdy = self.control.GetSizeTuple()
        pdx, pdy = self.popup.GetSizeTuple()
        pdx, pdy = max(pdx, cdx, self.width), max(pdy, self.height)

        # Calculate the best position and size for the pop-up:
        py = cy + cdy
        if (py + pdy) > screen_dy:
            if (cy - pdy) < 0:
                bottom = screen_dy - py
                if cy > bottom:
                    py, pdy = 0, cy
                else:
                    pdy = bottom
            else:
                py = cy - pdy

        # Finally, position the popup control:
        self.popup.SetDimensions(px, py, pdx, pdy)

    def _on_close_popup(self, event):
        """ Closes the popup control when it is deactivated.
        """
        if not event.GetActive():
            self._close_popup()

    def _close_popup(self):
        """ Closes the dialog.
        """
        wx.EVT_ACTIVATE(self.popup, None)
        self.dispose()
        self.closed = True
        self.popup.Destroy()
        self.popup = self.control = None

#-------------------------------------------------------------------------
#  'BufferDC' class:
#-------------------------------------------------------------------------


class BufferDC(wx.MemoryDC):
    """ An off-screen buffer class.

        This class implements a off-screen output buffer. Data is meant to
        be drawn in the buffer and then blitted directly to the output device
        context.
    """

    def __init__(self, dc, width=None, height=None):
        """Initializes the buffer."""
        wx.MemoryDC.__init__(self)

        # If only a single argument is passed, it is assumed to be a wx.Window
        # and that we have been created within a 'paint' event for that window:
        if width is None:
            width, height = dc.GetClientSize()
            dc = wx.PaintDC(dc)

        self.dc = dc
        self.bitmap = wx.EmptyBitmap(width, height)

        self.SelectObject(self.bitmap)

        self.SetFont(dc.GetFont())

    def copy(self, x=0, y=0):
        """ Performs the blit of the buffer contents to the specified device
            context location.
        """
        self.dc.Blit(x, y, self.bitmap.GetWidth(), self.bitmap.GetHeight(),
                     self, 0, 0)

#-------------------------------------------------------------------------
#  'Slider' class:
#-------------------------------------------------------------------------


class Slider(wx.Slider):
    """ This is a 'fixed' version of the wx.Slider control which does not
        erase its background, which can cause a lot of update flicker and is
        completely unnecessary.
    """

    def __init__(self, *args, **kw):
        super(Slider, self).__init__(*args, **kw)

        wx.EVT_ERASE_BACKGROUND(self, self._erase_background)

    def _erase_background(self, event):
        pass



"""

  fs.tests.test_opener:  testcases for FS opener

"""

import unittest
import tempfile
import shutil

from fs.opener import opener
from fs import path

class TestOpener(unittest.TestCase):

    def setUp(self):
        self.temp_dir = tempfile.mkdtemp(u"fstest_opener")

    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    def testOpen(self):
        filename = path.join(self.temp_dir, 'foo.txt')
        file_object = opener.open(filename, 'wb')
        file_object.close()
        self.assertTrue(file_object.closed)






